# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing the settings.
"""

from typing import Any
import os
from rich import print
import rich.table
import numpy as np

from mrmustard.utils.filters import add_complex_warning_filter, remove_complex_warning_filter

__all__ = ["Settings", "settings"]


class ImmutableSetting:
    r"""A setting that becomes immutable after the first time its value is queried.

    Args:
        value (any): the default value of this setting
        name (str): the name of this setting
    """

    def __init__(self, value: Any, name: str) -> None:
        self._value = value
        self._name = name
        self._is_immutable = False

    @property
    def name(self):
        r"""The name of this setting."""
        return self._name

    @property
    def value(self):
        r"""The value of this setting."""
        self._is_immutable = True
        return self._value

    @value.setter
    def value(self, value):
        if self._is_immutable:
            raise ValueError(f"Cannot change the value of `settings.{self.name}`.")
        self._value = value


# pylint: disable=too-many-instance-attributes
class Settings:
    r"""A class containing various settings that are used by Mr Mustard throughout a session.

    Some of these settings (such as those representing cutoff values) can be changed at any time,
    while others (such as the value of the Planck constant) can only be changed before being
    queried for the first time.

    .. code-block::

        from mrmustard import settings

        >>> settings.AUTOCUTOFF_MAX_CUTOFF  # check the default values
        100

        >>> settings.AUTOCUTOFF_MAX_CUTOFF = 150  # update to new values
        >>> settings.AUTOCUTOFF_MAX_CUTOFF
        150
    """

    def __new__(cls):  # singleton
        if not hasattr(cls, "instance"):
            cls.instance = super(Settings, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        self._hbar = ImmutableSetting(2.0, "HBAR")
        self._debug = False
        self._autocutoff_probability = 0.999  # capture at least 99.9% of the probability
        self._autocutoff_max_cutoff = 100
        self._autocutoff_min_cutoff = 1
        self._circuit_decimals = 3
        self._discretization_method = "iterative"
        # use cutoff=5 for each mode when determining if two transformations in fock repr are equal
        # 3 is enough to include a full step of the rec relations
        self._eq_transformation_cutoff = 3
        self._eq_transformation_rtol_fock = 1e-3
        self._eq_transformation_rtol_gauss = 1e-6
        # for the detectors
        self._pnr_internal_cutoff = 50
        self._homodyne_squeezing = 10.0
        # misc
        self._progressbar = True
        self._seed = np.random.randint(0, 2**31 - 1)
        self.rng = np.random.default_rng(self._seed)
        self._default_bs_method = "vanilla"  # can be 'vanilla' or 'schwinger'
        self._precision_bits_hermite_poly = 128
        self._julia_initialized = (
            False  # set to True when Julia is initialized (cf. PRECISION_BITS_HERMITE_POLY.setter)
        )
        self._complex_warning = False

    def _force_hbar(self, value):
        r"can set the value of HBAR at any time. use with caution."
        self._hbar._value = value

    @property
    def AUTOCUTOFF_MAX_CUTOFF(self):
        r"""The maximum value for autocutoff. Default is ``100``."""
        return self._autocutoff_max_cutoff

    @AUTOCUTOFF_MAX_CUTOFF.setter
    def AUTOCUTOFF_MAX_CUTOFF(self, value: int):
        self._autocutoff_max_cutoff = value

    @property
    def AUTOCUTOFF_MIN_CUTOFF(self):
        r"""The minimum value for autocutoff. Default is ``1``."""
        return self._autocutoff_min_cutoff

    @AUTOCUTOFF_MIN_CUTOFF.setter
    def AUTOCUTOFF_MIN_CUTOFF(self, value: int):
        self._autocutoff_min_cutoff = value

    @property
    def AUTOCUTOFF_PROBABILITY(self):
        r"""The autocutoff probability. Default is ``0.999``."""
        return self._autocutoff_probability

    @AUTOCUTOFF_PROBABILITY.setter
    def AUTOCUTOFF_PROBABILITY(self, value: float):
        self._autocutoff_probability = value

    @property
    def CIRCUIT_DECIMALS(self):
        r"""The number of decimals displayed when drawing a circuit with parameters. Default is ``3``."""
        return self._circuit_decimals

    @CIRCUIT_DECIMALS.setter
    def CIRCUIT_DECIMALS(self, value: int):
        self._circuit_decimals = value

    @property
    def COMPLEX_WARNING(self):
        r"""Whether tensorflow's ``ComplexWarning``s should be raised when a complex is casted to a float. Default is ``False``."""
        return self._complex_warning

    @COMPLEX_WARNING.setter
    def COMPLEX_WARNING(self, value: bool):
        self._complex_warning = value
        if value:
            remove_complex_warning_filter()
        else:
            add_complex_warning_filter()

    @property
    def DEBUG(self):
        r"""Whether or not to print the vector of means and the covariance matrix alongside the
        html representation of a state. Default is ``False``.
        """
        return self._debug

    @DEBUG.setter
    def DEBUG(self, value: bool):
        self._debug = value

    @property
    def DISCRETIZATION_METHOD(self):
        r"""The method used to discretize the Wigner function. Default is ``iterative``.

        Can be either ``'iterative'`` or ``'clenshaw'``.
        """
        return self._discretization_method

    @DISCRETIZATION_METHOD.setter
    def DISCRETIZATION_METHOD(self, value: str):
        self._discretization_method = value

    @property
    def DEFAULT_BS_METHOD(self):
        r"""The default method for computing the transformation operated by a beam splitter in
         the Fock basis . Default is ``vanilla``.

        Can be either ``'vanilla'`` or ``'schwinger'``.
        """
        return self._default_bs_method

    @DEFAULT_BS_METHOD.setter
    def DEFAULT_BS_METHOD(self, value: str):
        self._default_bs_method = value

    @property
    def EQ_TRANSFORMATION_CUTOFF(self):
        r"""The cutoff used when comparing two transformations via the Choi–Jamiolkowski
        isomorphism. Default is ``3``."""
        return self._eq_transformation_cutoff

    @EQ_TRANSFORMATION_CUTOFF.setter
    def EQ_TRANSFORMATION_CUTOFF(self, value: int):
        self._eq_transformation_cutoff = value

    @property
    def EQ_TRANSFORMATION_RTOL_FOCK(self):
        r"""The relative tolerance used when comparing two transformations via the Choi–Jamiolkowski
        isomorphism. Default is ``1e-3``."""
        return self._eq_transformation_rtol_fock

    @EQ_TRANSFORMATION_RTOL_FOCK.setter
    def EQ_TRANSFORMATION_RTOL_FOCK(self, value: float):
        self._eq_transformation_rtol_fock = value

    @property
    def EQ_TRANSFORMATION_RTOL_GAUSS(self):
        r"""The relative tolerance used when comparing two transformations on Gaussian states.
        Default is ``1e-6``."""
        return self._eq_transformation_rtol_gauss

    @EQ_TRANSFORMATION_RTOL_GAUSS.setter
    def EQ_TRANSFORMATION_RTOL_GAUSS(self, value: float):
        self._eq_transformation_rtol_gauss = value

    @property
    def HBAR(self):
        r"""The value of the Planck constant. Default is ``2``.

        Cannot be changed after its value is queried for the first time.
        """
        return self._hbar.value

    @HBAR.setter
    def HBAR(self, value: float):
        self._hbar.value = value

    @property
    def HOMODYNE_SQUEEZING(self):
        r"""The value of squeezing for homodyne measurements. Default is ``10``."""
        return self._homodyne_squeezing

    @HOMODYNE_SQUEEZING.setter
    def HOMODYNE_SQUEEZING(self, value: float):
        self._homodyne_squeezing = value

    @property
    def PNR_INTERNAL_CUTOFF(self):
        r"""The cutoff used when computing the output of a PNR detection. Default is ``50``."""
        return self._pnr_internal_cutoff

    @PNR_INTERNAL_CUTOFF.setter
    def PNR_INTERNAL_CUTOFF(self, value: int):
        self._pnr_internal_cutoff = value

    @property
    def PROGRESSBAR(self):
        r"""Whether or not to display the progress bar when performing training. Default is ``True``."""
        return self._progressbar

    @PROGRESSBAR.setter
    def PROGRESSBAR(self, value: bool):
        self._progressbar = value

    @property
    def SEED(self):
        r"""Returns the seed value if set, otherwise returns a random seed."""
        if self._seed is None:
            self._seed = np.random.randint(0, 2**31 - 1)
            self.rng = np.random.default_rng(self._seed)
        return self._seed

    @SEED.setter
    def SEED(self, value: int):
        self._seed = value
        self.rng = np.random.default_rng(self._seed)

    @property
    def PRECISION_BITS_HERMITE_POLY(self):
        r"""
        The number of bits used to represent a single Fock amplitude when calculating Hermite polynomials.
        Default is 128 (i.e. the Fock representation has dtype complex128).
        Currently allowed values: 128, 256, 384, 512
        """
        return self._precision_bits_hermite_poly

    @PRECISION_BITS_HERMITE_POLY.setter
    def PRECISION_BITS_HERMITE_POLY(self, value: int):
        allowed_values = [128, 256, 384, 512]
        if value not in allowed_values:
            raise ValueError(
                f"precision_bits_hermite_poly must be one of the following values: {allowed_values}"
            )
        self._precision_bits_hermite_poly = value
        if (
            value != 128 and not self._julia_initialized
        ):  # initialize Julia when precision > complex128 and if it wasn't initialized before
            from julia.api import LibJulia  # pylint: disable=import-outside-toplevel

            # the next line must be run before "from julia import Main as Main_julia"
            LibJulia.load().init_julia(
                ["--compiled-modules=no", "--project=julia_pkg"]
            )  # also loads julia environment
            # the next line must be run after "LibJulia.load().init_julia()"
            from julia import Main as Main_julia  # pylint: disable=import-outside-toplevel

            # import Julia functions
            utils_directory = os.path.dirname(__file__)
            Main_julia.cd(utils_directory)
            Main_julia.include("../math/lattice/strategies/julia/getPrecision.jl")
            Main_julia.include("../math/lattice/strategies/julia/vanilla.jl")
            Main_julia.include("../math/lattice/strategies/julia/compactFock/helperFunctions.jl")
            Main_julia.include("../math/lattice/strategies/julia/compactFock/diagonal_amps.jl")
            Main_julia.include("../math/lattice/strategies/julia/compactFock/diagonal_grad.jl")
            Main_julia.include(
                "../math/lattice/strategies/julia/compactFock/singleLeftoverMode_amps.jl"
            )
            Main_julia.include(
                "../math/lattice/strategies/julia/compactFock/singleLeftoverMode_grad.jl"
            )

            self._julia_initialized = True

    # use rich.table to print the settings
    def __repr__(self) -> str:
        r"""Returns a string representation of the settings."""

        # attributes that should not be displayed in the table
        not_displayed = ["rng"]

        table = rich.table.Table(title="MrMustard Settings")
        table.add_column("Setting")
        table.add_column("Value")

        for key, val in self.__dict__.items():
            if key in not_displayed:
                continue
            key = key.upper()[1:]
            value = str(val._value) if isinstance(val, ImmutableSetting) else str(val)
            table.add_row(key, value)

        print(table)
        return ""


settings = Settings()
"""Settings object."""



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of callback functionalities for optimizations.

Callbacks allow users to have finer control over the optimization process by executing
predefined routines as optimization progresses. Even though the :meth:`Optimizer.minimize` accepts
`Callable` functions directly, the :class:`Callback` class modularizes the logic and makes it
easier for users to inherit from it and come up with their own custom callbacks.

Things you can do with callbacks:

* Logging custom metrics.
* Tracking parameters and costs with Tensorboard.
* Scheduling learning rates.
* Modifying the gradient update that gets applied.
* Updating cost_fn to alter the optimization landscape in our favour.
* Adding some RL into the optimizer.
* ...

Builtin callbacks:

* :class:`Callback`: The base class, to be used for building custom callbacks.
* :class:`TensorboardCallback`: Tracks costs, parameter values and gradients in Tensorboard.

Examples:
=========

.. code-block::

    import numpy as np
    from mrmustard.training import Optimizer, TensorboardCallback

    def cost_fn():
        ...

    def as_dB(cost):
        delta = np.sqrt(np.log(1 / (abs(cost) ** 2)) / (2 * np.pi))
        cost_dB = -10 * np.log10(delta**2)
        return cost_dB

    tb_cb = TensorboardCallback(cost_converter=as_dB, track_grads=True)

    def rolling_cost_cb(optimizer, cost, **kwargs):
        return {
            'rolling_cost': np.mean(optimizer.opt_history[-10:] + [cost]),
        }

    opt = Optimizer(euclidean_lr = 0.001);
    opt.minimize(cost_fn, max_steps=200, by_optimizing=[...], callbacks=[tb_cb, rolling_cost_cb])

    # VScode can be used to open the Tensorboard frontend for live monitoring.

    opt.callback_history['TensorboardCallback']
    opt.callback_history['rolling_cost_cb']

"""

# pylint: disable = wrong-import-position


from dataclasses import dataclass
from datetime import datetime
import hashlib
from pathlib import Path
from typing import Callable, Optional, Mapping, Sequence, Union

import os
import numpy as np

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import tensorflow as tf

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "0"


@dataclass
class Callback:
    """Base callback class for optimizers. Users can inherit from this class and define the
    following custom logic:

    * `.trigger`:
        Custom triggering condition, other than the regular schedule set by `step_per_call`.
    * `.call`:
        The main routine to be customized.
    * `.update_cost_fn`:
        The custom cost_fn updater, which is expected to return a new cost_fn callable to
        replace the original one passed to the optimizer.
    * `.update_grads`:
        The custom grads modifyer, which is expected to return a list of parameter gradients
        after modification, to be applied to the parameters.
    * `.update_optimizer`:
        The custom optimizer updater, which is expected to modify the optimizer inplace for
        things like scheduling learning rates.

    """

    #: Custom tag for a callback instance to be used as keys in `Optimizer.callback_history`.
    #: Defaults to the class name.
    tag: str = None

    #: Sets calling frequency of this callback. Defaults to once per optimization step.
    #: Use higher values to reduce overhead.
    steps_per_call: int = 1

    def __post_init__(self):
        self.tag = self.tag or self.__class__.__name__
        self.optimizer_step: int = 0
        self.callback_step: int = 0

    def get_opt_step(self, optimizer, **kwargs):  # pylint: disable=unused-argument
        """Gets current step from optimizer."""
        self.optimizer_step = len(optimizer.opt_history)
        return self.optimizer_step

    def _should_call(self, **kwargs) -> bool:
        return (self.get_opt_step(**kwargs) % self.steps_per_call == 0) or self.trigger(**kwargs)

    def trigger(self, **kwargs) -> bool:  # pylint: disable=unused-argument
        """User implemented custom trigger conditions."""

    def call(self, **kwargs) -> Optional[Mapping]:  # pylint: disable=unused-argument
        """User implemented main callback logic."""

    def update_cost_fn(self, **kwargs) -> Optional[Callable]:  # pylint: disable=unused-argument
        """User implemented cost_fn modifier."""

    def update_grads(self, **kwargs) -> Optional[Sequence]:  # pylint: disable=unused-argument
        """User implemented gradient modifier."""

    def update_optimizer(self, optimizer, **kwargs):  # pylint: disable=unused-argument
        """User implemented optimizer update scheduler."""

    def __call__(
        self,
        **kwargs,
    ):
        if self._should_call(**kwargs):
            self.callback_step += 1
            callback_result = {
                "optimizer_step": self.optimizer_step,
                "callback_step": self.callback_step,
            }

            callback_result.update(self.call(**kwargs) or {})

            new_cost_fn = self.update_cost_fn(callback_result=callback_result, **kwargs)
            if callable(new_cost_fn):
                callback_result["cost_fn"] = new_cost_fn

            new_grads = self.update_grads(callback_result=callback_result, **kwargs)
            if new_grads is not None:
                callback_result["grads"] = new_grads

            # Modifies the optimizer inplace, e.g. its learning rates.
            self.update_optimizer(callback_result=callback_result, **kwargs)

            return callback_result
        return {}


@dataclass
class TensorboardCallback(Callback):  # pylint: disable=too-many-instance-attributes
    """Callback for enabling Tensorboard tracking of optimization progresses.

    Things tracked:

    * the cost
    * the transformed cost, if a `cost_converter` is provided
    * trainable parameter values
    * trainable parameter gradients (if `track_grads` is `True`)

    To start the Tensorboard frontend, either:

    * use VSCode: F1 -> Tensorboard -> select your `root_logdir/experiment_tag`.
    * use command line: `tensorboard --logdir=root_logdir/experiment_tag` and open link in browser.


    """

    #: The root logdir for tensorboard logging.
    root_logdir: Union[str, Path] = "./tb_logdir"

    #: The tag for experiment subfolder to group similar optimizations together for easy comparisons.
    #: Defaults to the hash of all trainable variables' names.
    experiment_tag: Optional[str] = None

    #: Extra prefix to name the optimization experiment.
    prefix: Optional[str] = None

    #: Transformation on cost for the purpose of better interpretation.
    cost_converter: Optional[Callable] = None

    #: Whether to track gradients as well as the values for trainable parameters.
    track_grads: bool = False

    #: Whether to return objectives in the callback results to be stored.
    log_objectives: bool = True

    #: Whether to return parameter values in the callback results to be stored.
    log_trainables: bool = False

    def __post_init__(self):
        super().__post_init__()
        self.root_logdir = Path(self.root_logdir)

        # Initialize only when first called to use optimization time rather than init time:
        self.logdir = None
        self.writter_logdir = None
        self.tb_writer = None

    def init_writer(self, trainables):
        """Initializes tb logdir folders and writer."""
        if (self.writter_logdir is None) or (self.optimizer_step <= self.steps_per_call):
            trainable_key_hash = hashlib.sha256(
                ",".join(trainables.keys()).encode("utf-8")
            ).hexdigest()
            self.experiment_tag = self.experiment_tag or f"experiment-{trainable_key_hash[:7]}"
            self.logdir = self.root_logdir / self.experiment_tag
            self.prefix = self.prefix or "optim"
            existing_exp = [path for path in self.logdir.glob("*") if path.is_dir()]
            optim_timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")

            self.writter_logdir = self.logdir / (
                f"{self.prefix}-{len(existing_exp):03d}-{optim_timestamp}"
            )
            self.tb_writer = tf.summary.create_file_writer(str(self.writter_logdir))
            self.tb_writer.set_as_default()

    def call(
        self,
        optimizer,
        cost,
        trainables,
        **kwargs,
    ):  # pylint: disable=unused-argument,arguments-differ
        """Logs costs and parameters to Tensorboard."""
        self.init_writer(trainables=trainables)
        obj_tag = "objectives"

        cost = np.array(cost).item()

        obj_scalars = {
            f"{obj_tag}/cost": cost,
        }
        if self.cost_converter is not None:
            obj_scalars[f"{obj_tag}/{self.cost_converter.__name__}(cost)"] = self.cost_converter(
                cost
            )

        if "orig_cost" in optimizer.callback_history:
            orig_cost = np.array(optimizer.callback_history["orig_cost"][-1]).item()
            obj_scalars[f"{obj_tag}/orig_cost"] = orig_cost
            if self.cost_converter is not None:
                obj_scalars[f"{obj_tag}/{self.cost_converter.__name__}(orig_cost)"] = (
                    self.cost_converter(orig_cost)
                )

        for k, v in obj_scalars.items():
            tf.summary.scalar(k, data=v, step=self.optimizer_step)

        for k, (x, dx) in trainables.items():
            x = np.array(x.value)
            if self.track_grads:
                dx = np.array(dx)

            tag = k if np.size(x) <= 1 else None
            for ind, val in np.ndenumerate(x):
                tag = tag or k + str(list(ind)).replace(" ", "")
                tf.summary.scalar(tag + ":value", data=val, step=self.optimizer_step)
                if self.track_grads:
                    tf.summary.scalar(tag + ":grad", data=dx[ind], step=self.optimizer_step)
                tag = None

        result = obj_scalars if self.log_objectives else {}

        if self.log_trainables:
            result.update(trainables)

        return result



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the calculation of the Wigner function."""

import numpy as np
from numba import njit

from mrmustard import math, settings


__all__ = ["wigner_discretized"]


# ~~~~~~~
# Helpers
# ~~~~~~~


@njit
def make_grid(q_vec, p_vec, hbar):  # pragma: no cover
    r"""Returns two coordinate matrices `Q` and `P` from coordinate vectors
    `q_vec` and `p_vec`, along with the grid over which Wigner functions can be
    discretized.
    """
    Q = np.outer(q_vec, np.ones_like(p_vec))
    P = np.outer(np.ones_like(q_vec), p_vec)
    return Q, P, (Q + P * 1.0j) / np.sqrt(2 * hbar)


@njit
def _wig_laguerre_val(L, x, diag):  # pragma: no cover
    """Returns the coefficient `c_L = sum_n rho_{n,L+n} Z_n^L` used
    by `_wigner_discretized_clenshaw`. The evaluation uses the Clenshaw recursion.
    """
    if len(diag) == 2:
        y0 = np.array([[diag[0]]], dtype=np.complex128)
        y1 = np.array([[diag[1]]], dtype=np.complex128)
    else:
        k = len(diag)
        y0 = np.array([[diag[-2]]], dtype=np.complex128)
        y1 = np.array([[diag[-1]]], dtype=np.complex128)
        for i in range(3, len(diag) + 1):
            k -= 1
            temp_y0 = y0
            y0 = diag[-i] - y1 * (float((k - 1) * (L + k - 1)) / ((L + k) * k)) ** 0.5
            y1 = temp_y0 - y1 * ((L + 2 * k - 1) - x) * ((L + k) * k) ** -0.5

    return y0 - y1 * ((L + 1) - x) * (L + 1) ** -0.5


# ~~~~~~~
# Methods
# ~~~~~~~


def wigner_discretized(rho, q_vec, p_vec):
    r"""Calculates the discretized Wigner function for a single mode.

    The supported discretization methods are:

    * ``iterative`` (default): Uses an iterative method to calculate the Wigner
    coefficients :math:`W_{mn}` in :math:`W = \sum_{mn} W_{mn} |m\rangle\langle n|`.
    This method is recommended for systems with low numbers of excitations (``n\leq50``).
    * ``clenshaw``: Uses Clenshaw summations to improve the performance for systems
    with large numbers of excitations (``n\leq50``).

    The discretization method can be changed by moodifying the `Settings` object.

    .. code::

        >>> settings.DISCRETIZATION_METHOD  # default method
        "iterative"

        >>> settings.DISCRETIZATION_METHOD = "clenshaw"  # change method

    These methods are adapted versions of the 'iterative' and 'clenshaw' methods of the
    Wigner function discretization routine provided in
    QuTiP <http://qutip.org/docs/4.0.2/apidoc/functions.html?highlight=wigner#qutip.wigner.wigner>`_,
    which is released under the BSD license, with the following copyright notice:

    Copyright (C) 2011 and later, P.D. Nation, J.R. Johansson,
    A.J.G. Pitchford, C. Granade, and A.L. Grimsmo. All rights reserved.

    Args:
        rho (complex array): the density matrix of the state in Fock representation
        q_vec (array): array of discretized :math:`q` quadrature values
        p_vec (array): array of discretized :math:`p` quadrature values

    Returns:
        tuple(array, array, array): array containing the discretized Wigner function, and the Q and
            P coordinates (in meshgrid form) in which the function is calculated
    """
    hbar = settings.HBAR
    method = settings.DISCRETIZATION_METHOD

    rho = math.asnumpy(rho)
    if method == "iterative":
        return _wigner_discretized_iterative(rho, q_vec, p_vec, hbar)
    return _wigner_discretized_clenshaw(rho, q_vec, p_vec, hbar)


@njit
def _wigner_discretized_clenshaw(rho, q_vec, p_vec, hbar):  # pragma: no cover
    r"""Calculates the Wigner function as
    :math:`W = C(x) \sum_L c_L (2x)^L / sqrt(L!)`, where:

    * :math:`x = (q + ip)`, for ``q`` and ``p`` in ``q_vec`` and ``p_vec``
      respectively
    * :math:`C(x) = e^{-x**2/(2\pi)}`
    * :math:`L` is the dimension of ``rho``
    * :math:`c_L = \sum_n \rho_{n,L+n} Z_n^L`
    * :math:`Z_n^L = (-1)^n sqrt(L!n!/(L+n)!) Lag(n,L,x)`
    * :math:`LaguerreL(n,L,x)`
    """
    cutoff = len(rho)
    Q, P, grid = make_grid(q_vec, p_vec, hbar)

    A = 2 * grid
    B = np.abs(A)
    B *= B

    w0 = (2 * rho[0, -1]) * np.ones_like(A)
    rho2 = rho * (2 * np.ones((cutoff, cutoff)) - np.diag(np.ones(cutoff)))

    L = cutoff - 1
    for j in range(1, cutoff):
        c_L = _wig_laguerre_val(L - j, B, np.diag(rho2, L - j))
        w0 = c_L + w0 * A * (L - j + 1) ** -0.5

    return w0.real * np.exp(-B * 0.5) / np.pi / hbar, Q, P


@njit
def _wigner_discretized_iterative(rho, q_vec, p_vec, hbar):  # pragma: no cover
    cutoff = len(rho)
    Q, P, grid = make_grid(q_vec, p_vec, hbar)
    Wmat = np.zeros((2, cutoff) + grid.shape, dtype=np.complex128)

    # W = rho(0,0)W(|0><0|)
    Wmat[0, 0] = np.exp(-2.0 * np.abs(grid) ** 2) / np.pi
    W = np.real(rho[0, 0]) * np.real(Wmat[0, 0])

    for n in range(1, cutoff):
        Wmat[0, n] = (2.0 * grid * Wmat[0, n - 1]) / np.sqrt(n)

        # W += rho(0,n)W(|0><n|) + rho(n,0)W(|n><0|)
        W += 2 * np.real(rho[0, n] * Wmat[0, n])

    for m in range(1, cutoff):
        Wmat[1, m] = (2 * np.conj(grid) * Wmat[0, m] - np.sqrt(m) * Wmat[0, m - 1]) / np.sqrt(m)

        # W = rho(m, m)W(|m><m|)
        W += np.real(rho[m, m] * Wmat[1, m])

        for n in range(m + 1, cutoff):
            Wmat[1, n] = (2 * grid * Wmat[1, n - 1] - np.sqrt(m) * Wmat[0, n - 1]) / np.sqrt(n)

            # W += rho(m,n)W(|m><n|) + rho(n,m)W(|n><m|)
            W += 2 * np.real(rho[m, n] * Wmat[1, n])
        Wmat[0] = Wmat[1]

    return W / hbar, Q, P



"""
This module calculates the derivatives for all possible Fock representations of mode 0, where all other modes are PNR detected.
This is done by applying the derivated recursion relation in a selective manner.
"""

import numpy as np
import numba
from numba import njit, int64
from numba.cpython.unsafe.tuple import tuple_setitem
from mrmustard.math.lattice.strategies.compactFock.helperFunctions import (
    SQRT,
    repeat_twice,
    construct_dict_params,
)


@njit
def calc_dA_dB(
    m,
    n,
    i,
    arr_read_pivot,
    read_GB,
    G_in_adapted,
    A_adapted,
    B,
    K_i,
    K_l_adapted,
    arr_read_pivot_dA,
    G_in_dA_adapted,
    arr_read_pivot_dB,
    G_in_dB_adapted,
    l_range,
):
    """
    Apply the derivated recurrence relation.
    """
    dA = arr_read_pivot_dA[(m, n) + read_GB] * B[i]
    dB = arr_read_pivot_dB[(m, n) + read_GB] * B[i]
    dB[i] += arr_read_pivot[(m, n) + read_GB]
    for l_prime, l in enumerate(l_range):
        dA += K_l_adapted[l_prime] * A_adapted[l_prime] * G_in_dA_adapted[l_prime]
        dB += K_l_adapted[l_prime] * A_adapted[l_prime] * G_in_dB_adapted[l_prime]
        dA[i, l] += G_in_adapted[l_prime]
    return dA / K_i[i - 2], dB / K_i[i - 2]


@njit
def write_block_grad(
    i,
    write,
    arr_read_pivot,
    read_GB,
    G_in,
    A,
    B,
    K_i,
    K_l,
    cutoff_leftoverMode,
    arr_write_dA,
    arr_read_pivot_dA,
    G_in_dA,
    arr_write_dB,
    arr_read_pivot_dB,
    G_in_dB,
):
    """
    Apply the derivated recurrence relation to blocks of Fock amplitudes (of shape cutoff_leftoverMode x cutoff_leftoverMode)
    This is the coarse-grained version of applying the derivated recurrence relation of mrmustard.math.compactFock.compactFock_diagonal_grad once.
    """
    # m,n = 0,0
    m, n = 0, 0
    l_range = np.arange(2, A.shape[1])
    A_adapted = A[i, 2:]
    G_in_adapted = G_in[0, 0]
    G_in_dA_adapted = G_in_dA[0, 0]
    G_in_dB_adapted = G_in_dB[0, 0]
    K_l_adapted = K_l
    arr_write_dA[(0, 0) + write], arr_write_dB[(0, 0) + write] = calc_dA_dB(
        m,
        n,
        i,
        arr_read_pivot,
        read_GB,
        G_in_adapted,
        A_adapted,
        B,
        K_i,
        K_l_adapted,
        arr_read_pivot_dA,
        G_in_dA_adapted,
        arr_read_pivot_dB,
        G_in_dB_adapted,
        l_range,
    )
    # m=0
    m = 0
    l_range = np.arange(1, A.shape[1])
    A_adapted = A[i, 1:]
    for n in range(1, cutoff_leftoverMode):
        K_l_adapted = np.hstack((np.array([np.sqrt(n)]), K_l))
        G_in_adapted = np.hstack(
            (np.array([arr_read_pivot[(0, n - 1) + read_GB] * np.sqrt(n)]), G_in[0, n])
        )
        G_in_dA_adapted = np.concatenate(
            (
                np.expand_dims(arr_read_pivot_dA[(0, n - 1) + read_GB], axis=0),
                G_in_dA[0, n],
            ),
            axis=0,
        )
        G_in_dB_adapted = np.concatenate(
            (
                np.expand_dims(arr_read_pivot_dB[(0, n - 1) + read_GB], axis=0),
                G_in_dB[0, n],
            ),
            axis=0,
        )
        arr_write_dA[(0, n) + write], arr_write_dB[(0, n) + write] = calc_dA_dB(
            m,
            n,
            i,
            arr_read_pivot,
            read_GB,
            G_in_adapted,
            A_adapted,
            B,
            K_i,
            K_l_adapted,
            arr_read_pivot_dA,
            G_in_dA_adapted,
            arr_read_pivot_dB,
            G_in_dB_adapted,
            l_range,
        )
    # n=0
    n = 0
    l_range = np.arange(1, A.shape[1])
    l_range[0] = 0
    A_adapted = np.hstack((np.array([A[i, 0]]), A[i, 2:]))
    for m in range(1, cutoff_leftoverMode):
        K_l_adapted = np.hstack((np.array([np.sqrt(m)]), K_l))
        G_in_adapted = np.hstack(
            (np.array([arr_read_pivot[(m - 1, 0) + read_GB] * np.sqrt(m)]), G_in[m, 0])
        )
        G_in_dA_adapted = np.concatenate(
            (
                np.expand_dims(arr_read_pivot_dA[(m - 1, 0) + read_GB], axis=0),
                G_in_dA[m, 0],
            ),
            axis=0,
        )
        G_in_dB_adapted = np.concatenate(
            (
                np.expand_dims(arr_read_pivot_dB[(m - 1, 0) + read_GB], axis=0),
                G_in_dB[m, 0],
            ),
            axis=0,
        )
        arr_write_dA[(m, 0) + write], arr_write_dB[(m, 0) + write] = calc_dA_dB(
            m,
            n,
            i,
            arr_read_pivot,
            read_GB,
            G_in_adapted,
            A_adapted,
            B,
            K_i,
            K_l_adapted,
            arr_read_pivot_dA,
            G_in_dA_adapted,
            arr_read_pivot_dB,
            G_in_dB_adapted,
            l_range,
        )
    # m>0,n>0
    l_range = np.arange(A.shape[1])
    A_adapted = A[i]
    for m in range(1, cutoff_leftoverMode):
        for n in range(1, cutoff_leftoverMode):
            K_l_adapted = np.hstack((np.array([np.sqrt(m), np.sqrt(n)]), K_l))
            G_in_adapted = np.hstack(
                (
                    np.array(
                        [
                            arr_read_pivot[(m - 1, n) + read_GB] * np.sqrt(m),
                            arr_read_pivot[(m, n - 1) + read_GB] * np.sqrt(n),
                        ]
                    ),
                    G_in[m, n],
                )
            )
            G_in_dA_adapted = np.concatenate(
                (
                    np.expand_dims(arr_read_pivot_dA[(m - 1, n) + read_GB], axis=0),
                    np.expand_dims(arr_read_pivot_dA[(m, n - 1) + read_GB], axis=0),
                    G_in_dA[m, n],
                ),
                axis=0,
            )
            G_in_dB_adapted = np.concatenate(
                (
                    np.expand_dims(arr_read_pivot_dB[(m - 1, n) + read_GB], axis=0),
                    np.expand_dims(arr_read_pivot_dB[(m, n - 1) + read_GB], axis=0),
                    G_in_dB[m, n],
                ),
                axis=0,
            )
            arr_write_dA[(m, n) + write], arr_write_dB[(m, n) + write] = calc_dA_dB(
                m,
                n,
                i,
                arr_read_pivot,
                read_GB,
                G_in_adapted,
                A_adapted,
                B,
                K_i,
                K_l_adapted,
                arr_read_pivot_dA,
                G_in_dA_adapted,
                arr_read_pivot_dB,
                G_in_dB_adapted,
                l_range,
            )
    return arr_write_dA, arr_write_dB


@njit
def read_block(
    arr_write,
    arr_write_dA,
    arr_write_dB,
    idx_write,
    arr_read,
    arr_read_dA,
    arr_read_dB,
    idx_read_tail,
    cutoff_leftoverMode,
):
    """
    Read the blocks of Fock amplitudes(of shape cutoff_leftoverMode x cutoff_leftoverMode)
    and their derivatives w.r.t A and B and write them to G_in, G_in_dA, G_in_dB
    """
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            arr_write[m, n, idx_write] = arr_read[
                (
                    m,
                    n,
                )
                + idx_read_tail
            ]
            arr_write_dA[m, n, idx_write] = arr_read_dA[
                (
                    m,
                    n,
                )
                + idx_read_tail
            ]
            arr_write_dB[m, n, idx_write] = arr_read_dB[
                (
                    m,
                    n,
                )
                + idx_read_tail
            ]

    return arr_write, arr_write_dA, arr_write_dB


@njit
def use_offDiag_pivot_grad(
    A,
    B,
    M,
    cutoff_leftoverMode,
    cutoffs_tail,
    params,
    d,
    arr0,
    arr2,
    arr1010,
    arr1001,
    arr1,
    arr0_dA,
    arr2_dA,
    arr1010_dA,
    arr1001_dA,
    arr1_dA,
    arr0_dB,
    arr2_dB,
    arr1010_dB,
    arr1001_dB,
    arr1_dB,
):
    """
    Apply recurrence relation for pivot of type [a+1,a,b,b,c,c,...] / [a,a,b+1,b,c,c,...] / [a,a,b,b,c+1,c,...]
    Args:
        A, B (array, Vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of detected modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        d (int): mode index in which the considered Fock amplitude is off diagonal
            e.g. [a,a,b+1,b,c,c,...] --> b is off diagonal --> d=1
        arr0, arr2, arr1010, arr1001, arr1 (array, array, array, array, array): submatrices of the fock representation
        arr..._dA, arr..._dB (array, array): derivatives of submatrices w.r.t A and B
    Returns:
        (array, array, array, array, array): updated versions of arr0, arr2, arr1010, arr1001, arr1
    """

    pivot = repeat_twice(params)
    pivot[2 * d] += 1
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, 2 * M), dtype=np.complex128)
    G_in_dA = np.zeros(G_in.shape + A.shape, dtype=np.complex128)
    G_in_dB = np.zeros(G_in.shape + B.shape, dtype=np.complex128)

    ########## READ ##########
    read_GB = (2 * d,) + params
    GB = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, len(B)), dtype=np.complex128)
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            GB[m, n] = arr1[(m, n) + read_GB] * B

    # Array0
    G_in, G_in_dA, G_in_dB = read_block(
        G_in,
        G_in_dA,
        G_in_dB,
        2 * d,
        arr0,
        arr0_dA,
        arr0_dB,
        params,
        cutoff_leftoverMode,
    )

    # read from Array2
    if params[d] > 0:
        params_adapted = tuple_setitem(params, d, params[d] - 1)
        G_in, G_in_dA, G_in_dB = read_block(
            G_in,
            G_in_dA,
            G_in_dB,
            2 * d + 1,
            arr2,
            arr2_dA,
            arr2_dB,
            (d,) + params_adapted,
            cutoff_leftoverMode,
        )

    # read from Array11
    for i in range(d + 1, M):  # i>d
        if params[i] > 0:
            params_adapted = tuple_setitem(params, i, params[i] - 1)
            G_in, G_in_dA, G_in_dB = read_block(
                G_in,
                G_in_dA,
                G_in_dB,
                2 * i,
                arr1001,
                arr1001_dA,
                arr1001_dB,
                (d, i - d - 1) + params_adapted,
                cutoff_leftoverMode,
            )
            G_in, G_in_dA, G_in_dB = read_block(
                G_in,
                G_in_dA,
                G_in_dB,
                2 * i + 1,
                arr1010,
                arr1010_dA,
                arr1010_dB,
                (d, i - d - 1) + params_adapted,
                cutoff_leftoverMode,
            )

    ########## WRITE ##########
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            G_in[m, n] = np.multiply(K_l, G_in[m, n])

    # Array0
    write = tuple_setitem(params, d, params[d] + 1)
    arr0_dA, arr0_dB = write_block_grad(
        2 * d + 3,
        write,
        arr1,
        read_GB,
        G_in,
        A,
        B,
        K_i,
        K_l,
        cutoff_leftoverMode,
        arr0_dA,
        arr1_dA,
        G_in_dA,
        arr0_dB,
        arr1_dB,
        G_in_dB,
    )

    # Array2
    if params[d] + 2 < cutoffs_tail[d]:
        write = (d,) + params
        arr2_dA, arr2_dB = write_block_grad(
            2 * d + 2,
            write,
            arr1,
            read_GB,
            G_in,
            A,
            B,
            K_i,
            K_l,
            cutoff_leftoverMode,
            arr2_dA,
            arr1_dA,
            G_in_dA,
            arr2_dB,
            arr1_dB,
            G_in_dB,
        )

    # Array11
    for i in range(d + 1, M):
        if params[i] + 1 < cutoffs_tail[i]:
            write = (d, i - d - 1) + params
            arr1010_dA, arr1010_dB = write_block_grad(
                2 * i + 2,
                write,
                arr1,
                read_GB,
                G_in,
                A,
                B,
                K_i,
                K_l,
                cutoff_leftoverMode,
                arr1010_dA,
                arr1_dA,
                G_in_dA,
                arr1010_dB,
                arr1_dB,
                G_in_dB,
            )
            arr1001_dA, arr1001_dB = write_block_grad(
                2 * i + 3,
                write,
                arr1,
                read_GB,
                G_in,
                A,
                B,
                K_i,
                K_l,
                cutoff_leftoverMode,
                arr1001_dA,
                arr1_dA,
                G_in_dA,
                arr1001_dB,
                arr1_dB,
                G_in_dB,
            )

    return (
        arr0_dA,
        arr2_dA,
        arr1010_dA,
        arr1001_dA,
        arr0_dB,
        arr2_dB,
        arr1010_dB,
        arr1001_dB,
    )


@njit
def use_diag_pivot_grad(
    A,
    B,
    M,
    cutoff_leftoverMode,
    cutoffs_tail,
    params,
    arr0,
    arr1,
    arr0_dA,
    arr1_dA,
    arr0_dB,
    arr1_dB,
):
    """
    Apply recurrence relation for pivot of type [a,a,b,b,c,c...]
    Args:
        A, B (array, Vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of detected modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        arr0, arr1 (array, array): submatrices of the fock representation
        arr..._dA, arr..._dB (array, array): derivatives of submatrices w.r.t A and B
    Returns:
        (array, array): updated versions of arr0, arr1
    """
    pivot = repeat_twice(params)
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, 2 * M), dtype=np.complex128)
    G_in_dA = np.zeros(G_in.shape + A.shape, dtype=np.complex128)
    G_in_dB = np.zeros(G_in.shape + B.shape, dtype=np.complex128)

    ########## READ ##########
    read_GB = params
    GB = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, len(B)), dtype=np.complex128)
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            GB[m, n] = arr0[(m, n) + read_GB] * B

    # Array1
    for i in range(2 * M):
        if params[i // 2] > 0:
            params_adapted = tuple_setitem(params, i // 2, params[i // 2] - 1)
            read = (
                i + 1 - 2 * (i % 2),
            ) + params_adapted  # [i+1-2*(i%2) for i in range(6)] == [1,0,3,2,5,4]
            G_in, G_in_dA, G_in_dB = read_block(
                G_in,
                G_in_dA,
                G_in_dB,
                i,
                arr1,
                arr1_dA,
                arr1_dB,
                read,
                cutoff_leftoverMode,
            )

    ########## WRITE ##########
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            G_in[m, n] = np.multiply(K_l, G_in[m, n])

    # Array1
    for i in range(2 * M):
        if params[i // 2] + 1 < cutoffs_tail[i // 2]:
            # this prevents a few elements from being written that will never be read
            if i != 1 or params[0] + 2 < cutoffs_tail[0]:
                write = (i,) + params
                arr1_dA, arr1_dB = write_block_grad(
                    i + 2,
                    write,
                    arr0,
                    read_GB,
                    G_in,
                    A,
                    B,
                    K_i,
                    K_l,
                    cutoff_leftoverMode,
                    arr1_dA,
                    arr0_dA,
                    G_in_dA,
                    arr1_dB,
                    arr0_dB,
                    G_in_dB,
                )

    return arr1_dA, arr1_dB


@njit
def fock_representation_1leftoverMode_grad_NUMBA(
    A,
    B,
    M,
    cutoff_leftoverMode,
    cutoffs_tail,
    arr0,
    arr2,
    arr1010,
    arr1001,
    arr1,
    tuple_type,
    list_type,
    zero_tuple,
):
    """
    Returns the gradients of the density matrices in the upper, undetected mode of a circuit when all other modes
    are PNR detected (according to algorithm 2 of https://doi.org/10.22331/q-2023-08-29-1097)
    Args:
        A, B (array, Vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        arr0 (array): submatrix of the fock representation that contains Fock amplitudes of the type [a,a,b,b,c,c...]
        arr2 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+2,a,b,b,c,c...] / [a,a,b+2,b,c,c...] / ...
        arr1010 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b+1,b,c,c,...] / [a+1,a,b,b,c+1,c,...] / [a,a,b+1,b,c+1,c,...] / ...
        arr1001 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b+1,c,c,...] / [a+1,a,b,b,c,c+1,...] / [a,a,b+1,b,c,c+1,...] / ...
        arr1 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b,c,c...] / [a,a+1,b,b,c,c...] / [a,a,b+1,b,c,c...] / ...
        tuple_type, list_type (numba types): numba types that need to be defined outside of numba compiled functions
    Returns:
        Tensor: the fock representation
    """
    arr0_dA = np.zeros(arr0.shape + A.shape, dtype=np.complex128)
    arr2_dA = np.zeros(arr2.shape + A.shape, dtype=np.complex128)
    arr1010_dA = np.zeros(arr1010.shape + A.shape, dtype=np.complex128)
    arr1001_dA = np.zeros(arr1001.shape + A.shape, dtype=np.complex128)
    arr1_dA = np.zeros(arr1.shape + A.shape, dtype=np.complex128)
    arr0_dB = np.zeros(arr0.shape + B.shape, dtype=np.complex128)
    arr2_dB = np.zeros(arr2.shape + B.shape, dtype=np.complex128)
    arr1010_dB = np.zeros(arr1010.shape + B.shape, dtype=np.complex128)
    arr1001_dB = np.zeros(arr1001.shape + B.shape, dtype=np.complex128)
    arr1_dB = np.zeros(arr1.shape + B.shape, dtype=np.complex128)

    # fill first mode for all PNR detections equal to zero
    for m in range(cutoff_leftoverMode - 1):
        arr0_dA[(m + 1, 0) + zero_tuple] = (
            arr0_dA[(m, 0) + zero_tuple] * B[0]
            + np.sqrt(m) * A[0, 0] * arr0_dA[(m - 1, 0) + zero_tuple]
        ) / np.sqrt(m + 1)
        arr0_dA[(m + 1, 0) + zero_tuple][0, 0] += (
            np.sqrt(m) * arr0[(m - 1, 0) + zero_tuple]
        ) / np.sqrt(m + 1)
        arr0_dB[(m + 1, 0) + zero_tuple] = (
            arr0_dB[(m, 0) + zero_tuple] * B[0]
            + np.sqrt(m) * A[0, 0] * arr0_dB[(m - 1, 0) + zero_tuple]
        ) / np.sqrt(m + 1)
        arr0_dB[(m + 1, 0) + zero_tuple][0] += arr0[(m, 0) + zero_tuple] / np.sqrt(m + 1)

    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode - 1):
            arr0_dA[(m, n + 1) + zero_tuple] = (
                arr0_dA[(m, n) + zero_tuple] * B[1]
                + np.sqrt(m) * A[1, 0] * arr0_dA[(m - 1, n) + zero_tuple]
                + np.sqrt(n) * A[1, 1] * arr0_dA[(m, n - 1) + zero_tuple]
            ) / np.sqrt(n + 1)
            arr0_dA[(m, n + 1) + zero_tuple][1, 0] += (
                np.sqrt(m) * arr0[(m - 1, n) + zero_tuple]
            ) / np.sqrt(n + 1)
            arr0_dA[(m, n + 1) + zero_tuple][1, 1] += (
                np.sqrt(n) * arr0[(m, n - 1) + zero_tuple]
            ) / np.sqrt(n + 1)
            arr0_dB[(m, n + 1) + zero_tuple] = (
                arr0_dB[(m, n) + zero_tuple] * B[1]
                + np.sqrt(m) * A[1, 0] * arr0_dB[(m - 1, n) + zero_tuple]
                + np.sqrt(n) * A[1, 1] * arr0_dB[(m, n - 1) + zero_tuple]
            ) / np.sqrt(n + 1)
            arr0_dB[(m, n + 1) + zero_tuple][1] += arr0[(m, n) + zero_tuple] / np.sqrt(n + 1)

    dict_params = construct_dict_params(cutoffs_tail, tuple_type, list_type)
    for sum_params in range(sum(cutoffs_tail)):
        for params in dict_params[sum_params]:
            # diagonal pivots: aa,bb,cc,dd,...
            if (cutoffs_tail[0] == 1) or (params[0] < cutoffs_tail[0] - 1):
                arr1_dA, arr1_dB = use_diag_pivot_grad(
                    A,
                    B,
                    M - 1,
                    cutoff_leftoverMode,
                    cutoffs_tail,
                    params,
                    arr0,
                    arr1,
                    arr0_dA,
                    arr1_dA,
                    arr0_dB,
                    arr1_dB,
                )
            # off-diagonal pivots: d=0: (a+1)a,bb,cc,dd,... | d=1: 00,(b+1)b,cc,dd | 00,00,(c+1)c,dd | ...
            for d in range(M - 1):
                if np.all(np.array(params)[:d] == 0) and (params[d] < cutoffs_tail[d] - 1):
                    (
                        arr0_dA,
                        arr2_dA,
                        arr1010_dA,
                        arr1001_dA,
                        arr0_dB,
                        arr2_dB,
                        arr1010_dB,
                        arr1001_dB,
                    ) = use_offDiag_pivot_grad(
                        A,
                        B,
                        M - 1,
                        cutoff_leftoverMode,
                        cutoffs_tail,
                        params,
                        d,
                        arr0,
                        arr2,
                        arr1010,
                        arr1001,
                        arr1,
                        arr0_dA,
                        arr2_dA,
                        arr1010_dA,
                        arr1001_dA,
                        arr1_dA,
                        arr0_dB,
                        arr2_dB,
                        arr1010_dB,
                        arr1001_dB,
                        arr1_dB,
                    )
    return arr0_dA, arr0_dB


def fock_representation_1leftoverMode_grad(A, B, M, arr0, arr2, arr1010, arr1001, arr1):
    """
    First initialise the submatrices of G (of which the shape depends on cutoff and M)
    and some other constants
    (These initialisations currently cannot be done using Numba.)
    Then calculate the fock representation.
    """

    cutoffs = tuple(arr0.shape[1:])
    cutoff_leftoverMode = cutoffs[0]
    cutoffs_tail = tuple(cutoffs[1:])
    tuple_type = numba.types.UniTuple(int64, M - 1)
    list_type = numba.types.ListType(tuple_type)
    zero_tuple = (0,) * (M - 1)

    return fock_representation_1leftoverMode_grad_NUMBA(
        A,
        B,
        M,
        cutoff_leftoverMode,
        cutoffs_tail,
        arr0,
        arr2,
        arr1010,
        arr1001,
        arr1,
        tuple_type,
        list_type,
        zero_tuple,
    )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

" This module contains strategies for calculating the matrix elements of the displacement gate. "

import numpy as np
from numba import jit

__all__ = ["displacement", "laguerre", "grad_displacement", "jacobian_displacement"]


@jit(nopython=True)
def displacement(cutoffs, alpha, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the matrix elements of the displacement gate using a recurrence relation.
    Uses the log of the matrix elements to avoid numerical issues and then takes the exponential.

    Args:
        alpha (complex): displacement magnitude and angle
        cutoffs (tuple[int, int]): Fock ladder output-input cutoffs
        dtype (data type): Specifies the data type used for the calculation

    Returns:
        array[complex]: matrix representing the displacement operation.
    """
    r = np.abs(alpha)
    phi = np.angle(alpha)
    N, M = cutoffs
    flipped = False
    if N < M:
        N, M = M, N
        flipped = True

    D = np.zeros((N, M), dtype=dtype)
    rng = np.arange(max(*cutoffs))
    rng[0] = 1
    log_k_fac = np.cumsum(np.log(rng))
    for n_minus_m in range(N):
        m_max = min(M, N - n_minus_m)
        logL = np.log(laguerre(r**2.0, m_max, n_minus_m))
        for m in range(m_max):
            n = n_minus_m + m
            sign = 2 * (not (flipped and n > m and n_minus_m % 2)) - 1
            conj = 2 * (not (flipped and n > m)) - 1
            D[n, m] = sign * np.exp(
                +0.5 * (log_k_fac[m] - log_k_fac[n])
                + n_minus_m * np.log(r)
                - (r**2.0) / 2.0
                + conj * 1j * phi * n_minus_m
                + logL[m]
            )
            if n < M:
                D[m, n] = (-1.0) ** n_minus_m * np.conj(D[n, m])
    return D if not flipped else np.transpose(D)


@jit(nopython=True, cache=True)
def laguerre(x, N, alpha, dtype=np.complex128):  # pragma: no cover
    r"""Returns the N first generalized Laguerre polynomials evaluated at x.

    Args:
        x (float): point at which to evaluate the polynomials
        N (int): maximum Laguerre polynomial to calculate
        alpha (float): continuous parameter for the generalized Laguerre polynomials
    """
    L = np.zeros(N, dtype=dtype)
    L[0] = 1.0
    if N > 1:
        for m in range(0, N - 1):
            L[m + 1] = ((2 * m + 1 + alpha - x) * L[m] - (m + alpha) * L[m - 1]) / (m + 1)
    return L


@jit(nopython=True)
def grad_displacement(T, r, phi):  # pragma: no cover
    r"""Calculates the gradient of the displacement gate with respect to the magnitude and angle of the displacement.

    Args:
        T (array[complex]): array representing the gate
        r (float): displacement magnitude
        phi (float): displacement angle

    Returns:
        tuple[array[complex], array[complex]]: The gradient of the displacement gate with respect to r and phi
    """
    cutoff = T.shape[0]
    dtype = T.dtype
    ei = np.exp(1j * phi)
    eic = np.exp(-1j * phi)
    alpha = r * ei
    alphac = r * eic
    sqrt = np.sqrt(np.arange(cutoff, dtype=dtype))
    grad_r = np.zeros((cutoff, cutoff), dtype=dtype)
    grad_phi = np.zeros((cutoff, cutoff), dtype=dtype)

    for m in range(cutoff):
        for n in range(cutoff):
            grad_r[m, n] = -r * T[m, n] + sqrt[m] * ei * T[m - 1, n] - sqrt[n] * eic * T[m, n - 1]
            grad_phi[m, n] = (
                sqrt[m] * 1j * alpha * T[m - 1, n] + sqrt[n] * 1j * alphac * T[m, n - 1]
            )

    return grad_r, grad_phi


@jit(nopython=True)
def jacobian_displacement(D, alpha):  # pragma: no cover
    r"""Calculates the jacobian of the displacement gate with respect to the complex displacement
    alpha and its conjugate. Both are needed for backprop, as the displacement gate is not a
    holomorphic function, as the Fock amplitudes depend on both alpha and on its conjugate.
    Each jacobian in this case has the same shape as the array D, as the displacement is a scalar.

    Args:
        D (array[complex]): the D(alpha) gate in Fock representation. Batch dimensions are allowed.
        alpha (complex): parameter of D(alpha)

    Returns:
        2 array[complex]: The jacobian of the displacement gate with respect to alpha and alphaconj
    """
    shape = D.shape[-2:]
    alphac = np.conj(alpha)
    sqrt = np.sqrt(np.arange(shape[0] + shape[1], dtype=D.dtype))
    jac_alpha = np.zeros(D.shape, dtype=D.dtype)  # i.e. dD_dalpha for all m,n
    jac_alphac = np.zeros(D.shape, dtype=D.dtype)  # i.e. dD_dalphac for all m,n
    for m in range(shape[0]):
        for n in range(shape[1]):
            jac_alpha[m, n] = -0.5 * alphac * D[m, n] + sqrt[m] * D[m - 1, n]
            jac_alphac[m, n] = -0.5 * alpha * D[m, n] - sqrt[n] * D[m, n - 1]
    return jac_alpha, jac_alphac



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements the set of detector classes that perform measurements on quantum circuits.
"""

from typing import Iterable, List, Optional, Tuple, Union

from mrmustard import settings
from mrmustard.physics import fock, gaussian
from mrmustard.utils.typing import RealMatrix, RealVector

from mrmustard import math
from .abstract import FockMeasurement, Measurement, State
from .gates import Rgate
from .states import Coherent, DisplacedSqueezed
from .utils import make_parameter

__all__ = ["PNRDetector", "ThresholdDetector", "Generaldyne", "Homodyne", "Heterodyne"]


# pylint: disable=no-member
class PNRDetector(FockMeasurement):
    r"""Photon Number Resolving detector.

    If ``len(modes) > 1`` the detector is applied in parallel to all of the modes provided.
    If a parameter is a single float, the parallel instances of the detector share that parameter.

    To apply mode-specific parmeters use a list of floats. The number of modes is determined (in order of priority)
    by the modes parameter, the cutoffs parameter, or the length of the efficiency and dark counts parameters.

    One can optionally set bounds for each parameter, which the optimizer will respect.

    It can be supplied the full stochastic channel, or it will compute it from
    the quantum efficiency (binomial) and the dark count probability (possonian).

    Args:
        efficiency (float or List[float]): list of quantum efficiencies for each detector
        efficiency_trainable (bool): whether the efficiency is trainable
        efficiency_bounds (Tuple[float, float]): bounds for the efficiency
        dark_counts (float or List[float]): list of expected dark counts
        dark_counts_trainable (bool): whether the dark counts are trainable
        dark_counts_bounds (Tuple[float, float]): bounds for the dark counts
        stochastic_channel (Optional 2d array): if supplied, this stochastic_channel will be used for belief propagation
        modes (Optional List[int]): list of modes to apply the detector to
        cutoffs (int or List[int]): largest phton number measurement cutoff for each mode
    """

    def __init__(
        self,
        efficiency: Union[float, List[float]] = 1.0,
        dark_counts: Union[float, List[float]] = 0.0,
        efficiency_trainable: bool = False,
        dark_counts_trainable: bool = False,
        efficiency_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        dark_counts_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        stochastic_channel: RealMatrix = None,
        modes: List[int] = None,
        cutoffs: Union[int, List[int]] = None,
    ):
        self._stochastic_channel = stochastic_channel
        self._should_recompute_stochastic_channel = efficiency_trainable or dark_counts_trainable

        eff = math.atleast_1d(efficiency)
        dk = math.atleast_1d(dark_counts)

        if modes is not None:
            num_modes = len(modes)
        elif cutoffs is not None:
            num_modes = len(cutoffs)
        else:
            num_modes = max(len(eff), len(dk))

        modes = modes or list(range(num_modes))
        outcome = None
        super().__init__(outcome, modes, cutoffs)

        self._add_parameter(
            make_parameter(efficiency_trainable, eff, "efficiency", efficiency_bounds)
        )
        self._add_parameter(
            make_parameter(dark_counts_trainable, dk, "dark_counts", dark_counts_bounds)
        )

        self.recompute_stochastic_channel()

    def should_recompute_stochastic_channel(self):
        return self._should_recompute_stochastic_channel

    def recompute_stochastic_channel(self, cutoffs: List[int] = None):
        """recompute belief using the defined `stochastic channel`"""
        if cutoffs is None:
            cutoffs = [settings.PNR_INTERNAL_CUTOFF] * len(self._modes)
        self._internal_stochastic_channel = []
        if self._stochastic_channel is not None:
            self._internal_stochastic_channel = self._stochastic_channel
        else:
            efficiency = (
                math.tile(math.atleast_1d(self.efficiency.value), [len(cutoffs)])
                if len(math.atleast_1d(self.efficiency.value)) == 1
                else self.efficiency.value
            )
            dark_counts = (
                math.tile(math.atleast_1d(self.dark_counts.value), [len(cutoffs)])
                if len(math.atleast_1d(self.dark_counts.value)) == 1
                else self.dark_counts.value
            )
            for c, qe, dc in zip(cutoffs, efficiency, dark_counts):
                dark_prior = math.poisson(max_k=settings.PNR_INTERNAL_CUTOFF, rate=dc)
                condprob = math.binomial_conditional_prob(
                    success_prob=qe, dim_in=settings.PNR_INTERNAL_CUTOFF, dim_out=c
                )
                self._internal_stochastic_channel.append(
                    math.convolve_probs_1d(
                        condprob, [dark_prior, math.eye(settings.PNR_INTERNAL_CUTOFF)[0]]
                    )
                )


# pylint: disable: no-member
class ThresholdDetector(FockMeasurement):
    r"""Threshold detector: any Fock component other than vacuum counts toward a click in the detector.

    If ``len(modes) > 1`` the detector is applied in parallel to all of the modes provided.

    If a parameter is a single float, its value is applied to all of the parallel instances of the detector.

    To apply mode-specific values use a list of floats.

    It can be supplied the full conditional detection probabilities, or it will compute them from
    the quantum efficiency (binomial) and the dark count probability (bernoulli).

    Args:
        efficiency (float or List[float]): list of quantum efficiencies for each detector
        dark_count_prob (float or List[float]): list of dark count probabilities for each detector
        efficiency_trainable (bool): whether the efficiency is trainable
        dark_count_prob_trainable (bool): whether the dark count probabilities are trainable
        efficiency_bounds (Tuple[float, float]): bounds for the efficiency
        dark_count_prob_bounds (Tuple[float, float]): bounds for the dark count probabilities
        stochastic_channel (Optional 2d array): if supplied, this stochastic_channel will be used for belief propagation
        modes (Optional List[int]): list of modes to apply the detector to
    """

    def __init__(
        self,
        efficiency: Union[float, List[float]] = 1.0,
        dark_count_prob: Union[float, List[float]] = 0.0,
        efficiency_trainable: bool = False,
        dark_count_prob_trainable: bool = False,
        efficiency_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        dark_count_prob_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        stochastic_channel=None,
        modes: List[int] = None,
    ):
        if modes is not None:
            num_modes = len(modes)
        else:
            num_modes = max(len(math.atleast_1d(efficiency)), len(math.atleast_1d(dark_count_prob)))

        if len(math.atleast_1d(efficiency)) == 1 and num_modes > 1:
            efficiency = math.tile(math.atleast_1d(efficiency), [num_modes])
        if len(math.atleast_1d(dark_count_prob)) == 1 and num_modes > 1:
            dark_count_prob = math.tile(math.atleast_1d(dark_count_prob), [num_modes])

        modes = modes or list(range(num_modes))

        self._stochastic_channel = stochastic_channel

        cutoffs = [2] * num_modes
        self._should_recompute_stochastic_channel = (
            efficiency_trainable or dark_count_prob_trainable
        )

        outcome = None
        super().__init__(outcome, modes, cutoffs)

        self._add_parameter(
            make_parameter(efficiency_trainable, efficiency, "efficiency", efficiency_bounds)
        )
        self._add_parameter(
            make_parameter(
                dark_count_prob_trainable,
                dark_count_prob,
                "dark_count_prob",
                dark_count_prob_bounds,
            )
        )

        self.recompute_stochastic_channel()

    def should_recompute_stochastic_channel(self):
        return self._should_recompute_stochastic_channel

    def recompute_stochastic_channel(self, cutoffs: List[int] = None):
        """recompute belief using the defined `stochastic channel`"""
        if cutoffs is None:
            cutoffs = [settings.PNR_INTERNAL_CUTOFF] * len(self._modes)
        self._internal_stochastic_channel = []
        if self._stochastic_channel is not None:
            self._internal_stochastic_channel = self._stochastic_channel
        else:
            for cut, qe, dc in zip(
                cutoffs,
                math.atleast_1d(self.efficiency.value)[:],
                math.atleast_1d(self.dark_count_prob.value)[:],
            ):
                row1 = math.pow(1.0 - qe, math.arange(cut)[None, :]) - math.cast(
                    dc, self.efficiency.value.dtype
                )
                row2 = 1.0 - row1
                # rest = math.zeros((cut - 2, cut), dtype=row1.dtype)
                condprob = math.concat([row1, row2], axis=0)
                self._internal_stochastic_channel.append(condprob)


class Generaldyne(Measurement):
    r"""Generaldyne measurement on given modes.

    Args:
        state (State): the Gaussian state of the measurment device
        outcome (optional or List[float]): the means of the measurement state, defaults to ``None``
        modes (List[int]): the modes on which the measurement is acting on
    """

    def __init__(
        self,
        state: State,
        outcome: Optional[RealVector] = None,
        modes: Optional[Iterable[int]] = None,
    ) -> None:
        if not state.is_gaussian:
            raise TypeError("Generaldyne measurement state must be Gaussian.")
        if outcome is not None and not outcome.shape == state.means.shape:
            raise TypeError(
                f"Expected `outcome` of size {state.means.shape} but got {outcome.shape}."
            )

        self.state = state

        if modes is None:
            modes = self.state.modes
        else:
            # ensure measurement and state act on the same modes
            self.state = state[modes]

        super().__init__(outcome, modes)

    @property
    def outcome(self) -> RealVector:
        return self.state.means

    def primal(self, other: State) -> Union[State, float]:
        if self.postselected:
            # return the projection of self.state onto other
            return self.state.primal(other)

        return super().primal(other)

    def _measure_gaussian(self, other) -> Union[State, float]:
        remaining_modes = list(set(other.modes) - set(self.modes))

        outcome, prob, new_cov, new_means = gaussian.general_dyne(
            other.cov, other.means, self.state.cov, None, modes=self.modes
        )
        self.state = State(cov=self.state.cov, means=outcome)

        return (
            prob
            if len(remaining_modes) == 0
            else State(cov=new_cov, means=new_means, modes=remaining_modes, _norm=prob)
        )

    def _measure_fock(self, other) -> Union[State, float]:
        raise NotImplementedError(f"Fock sampling not implemented for {self.__class__.__name__}")


class Heterodyne(Generaldyne):
    r"""Heterodyne measurement on given modes.

    This class is just a thin wrapper around the :class:`Coherent`.
    If neither ``x`` or ``y`` is provided then values will be sampled.

    Args:
        x (optional float or List[float]): the x-displacement of the coherent state, defaults to ``None``
        y (optional or List[float]): the y-displacement of the coherent state, defaults to ``None``
        modes (List[int]): the modes of the coherent state
    """

    def __init__(
        self,
        x: Union[float, List[float]] = 0.0,
        y: Union[float, List[float]] = 0.0,
        modes: List[int] = None,
    ):
        if (x is None) ^ (y is None):  # XOR
            raise ValueError("Both `x` and `y` arguments should be defined or set to `None`.")

        # if no x and y provided, sample the outcome
        if x is None and y is None:
            num_modes = len(modes) if modes is not None else 1
            x, y = math.zeros([num_modes]), math.zeros([num_modes])
            outcome = None
        else:
            x = math.atleast_1d(x, dtype="float64")
            y = math.atleast_1d(y, dtype="float64")
            outcome = math.concat([x, y], axis=0)  # XXPP ordering

        modes = modes or list(range(x.shape[0]))

        units_factor = math.sqrt(2.0 * settings.HBAR, dtype="float64")
        state = Coherent(x / units_factor, y / units_factor)
        super().__init__(state=state, outcome=outcome, modes=modes)


class Homodyne(Generaldyne):
    """Homodyne measurement on given modes. If ``result`` is not provided then the value
    is sampled.

    Args:
        quadrature_angle (float or List[float]): measurement quadrature angle
        result (optional float or List[float]): displacement amount
        modes (optional List[int]): the modes of the displaced squeezed state
        r (optional float or List[float]): squeezing amount (default: ``settings.HOMODYNE_SQUEEZING``)
    """

    def __init__(
        self,
        quadrature_angle: Union[float, List[float]],
        result: Optional[Union[float, List[float]]] = None,
        modes: Optional[List[int]] = None,
        r: Optional[Union[float, List[float]]] = None,
    ):
        self.r = r or settings.HOMODYNE_SQUEEZING
        self.quadrature_angle = math.atleast_1d(quadrature_angle, dtype="float64")

        # if no ``result`` provided, sample the outcome
        if result is None:
            x = math.zeros_like(self.quadrature_angle)
            y = math.zeros_like(self.quadrature_angle)
            outcome = None
        else:
            result = math.atleast_1d(result, dtype="float64")
            if result.shape[-1] == 1:
                result = math.tile(result, self.quadrature_angle.shape)

            x = result * math.cos(self.quadrature_angle)
            y = result * math.sin(self.quadrature_angle)
            outcome = math.concat([x, y], axis=0)  # XXPP ordering

        modes = modes or list(range(self.quadrature_angle.shape[0]))

        units_factor = math.sqrt(2.0 * settings.HBAR, dtype="float64")
        state = DisplacedSqueezed(
            r=self.r, phi=2 * self.quadrature_angle, x=x / units_factor, y=y / units_factor
        )
        super().__init__(state=state, outcome=outcome, modes=modes)

    def _measure_gaussian(self, other) -> Union[State, float]:
        # rotate modes to be measured to the Homodyne basis
        other >>= Rgate(-self.quadrature_angle, modes=self.modes)
        self.state >>= Rgate(-self.quadrature_angle, modes=self.modes)

        # perform homodyne measurement as a generaldyne one
        out = super()._measure_gaussian(other)

        # set p-outcomes to 0 and rotate the measurement device state back to the original basis,
        # this is in turn rotating the outcomes to the original basis
        self_state_means = math.concat(
            [self.state.means[: self.num_modes], math.zeros((self.num_modes,))], axis=0
        )
        self.state = State(cov=self.state.cov, means=self_state_means, modes=self.modes) >> Rgate(
            self.quadrature_angle, modes=self.modes
        )

        return out

    def _measure_fock(self, other) -> Union[State, float]:
        if len(self.modes) > 1:
            raise NotImplementedError(
                "Multimode Homodyne sampling for Fock representation is not yet implemented."
            )

        other_cutoffs = [
            None if m not in self.modes else other.cutoffs[other.indices(m)] for m in other.modes
        ]
        remaining_modes = list(set(other.modes) - set(self.modes))

        # create reduced state of modes to be measured on the homodyne basis
        reduced_state = other.get_modes(self.modes)

        # build pdf and sample homodyne outcome
        x_outcome, probability = fock.sample_homodyne(
            state=reduced_state.ket() if reduced_state.is_pure else reduced_state.dm(),
            quadrature_angle=self.quadrature_angle,
        )

        # Define conditional state of the homodyne measurement device and rotate back to the original basis.
        # Note: x_outcome already has units of sqrt(hbar). Here is divided by sqrt(2*hbar) to cancel the multiplication
        # factor of the displacement symplectic inside the DisplacedSqueezed state.
        x_arg = x_outcome / math.sqrt(2.0 * settings.HBAR, dtype="float64")
        self.state = DisplacedSqueezed(
            r=self.r, phi=0.0, x=x_arg, y=0.0, modes=self.modes
        ) >> Rgate(self.quadrature_angle, modes=self.modes)

        if remaining_modes == 0:
            return probability

        self_cutoffs = [other.cutoffs[other.indices(m)] for m in self.modes]
        other_cutoffs = [
            None if m not in self.modes else other.cutoffs[other.indices(m)] for m in other.modes
        ]
        out_fock = fock.contract_states(
            stateA=other.ket(other_cutoffs) if other.is_pure else other.dm(other_cutoffs),
            stateB=self.state.ket(self_cutoffs),
            a_is_dm=other.is_mixed,
            b_is_dm=False,
            modes=other.indices(self.modes),
            normalize=False,
        )

        return (
            State(dm=out_fock, modes=remaining_modes, _norm=probability)
            if other.is_mixed
            else State(ket=out_fock, modes=remaining_modes, _norm=probability)
        )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains a class to filter undesired warnings.
"""

import logging


class WarningFilters(logging.Filter):
    r"""
    A custom logging filter to selectively allow log records based on specific warnings.

    Args:
        warnings: A list of warning messages that must be filtered.
    """

    def __init__(self, warnings: list[str]):
        super().__init__()
        self.warnings = warnings

    def filter(self, record) -> bool:
        r"""
        Determine if the log record should be allowed based on specific warnings.

        Args:
            record: The ``LogRecord`` to be filtered.

        Returns:
            ``True`` if the log record should be allowed, ``False`` otherwise.
        """
        return any(w in record.getMessage() for w in self.warnings)


# ComplexWarning filter for tensorflow.
msg = "WARNING:tensorflow:You are casting an input of type complex128 to an incompatible dtype float64."
msg += "  This will discard the imaginary part and may not be what you intended."
complex_warninig_filter = WarningFilters([msg])


def add_complex_warning_filter():
    r"""
    Adds the filter for tensorflow's ComplexWarning, or does nothing if the filter is already in place.
    """
    logger = logging.getLogger("tensorflow")
    logger.addFilter(complex_warninig_filter)


def remove_complex_warning_filter():
    r"""
    Removes the filter for tensorflow's ComplexWarning, or does nothing if no such filter is present.
    """
    logger = logging.getLogger("tensorflow")
    logger.removeFilter(complex_warninig_filter)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the numpy backend."""

# pylint: disable = missing-function-docstring, missing-class-docstring, fixme


from math import lgamma as mlgamma
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import scipy as sp
from scipy.linalg import expm as scipy_expm
from scipy.linalg import sqrtm as scipy_sqrtm
from scipy.special import xlogy as scipy_xlogy
from scipy.stats import multivariate_normal

from ..utils.settings import settings
from .autocast import Autocast
from .backend_base import BackendBase
from .lattice.strategies import binomial, vanilla, vanilla_batch
from .lattice.strategies.compactFock.inputValidation import (
    hermite_multidimensional_1leftoverMode,
    hermite_multidimensional_diagonal,
    hermite_multidimensional_diagonal_batch,
)


# pylint: disable=too-many-public-methods
class BackendNumpy(BackendBase):  # pragma: no cover
    r"""
    A numpy backend.
    """

    int32 = np.int32
    float32 = np.float32
    float64 = np.float64
    complex64 = np.complex64
    complex128 = np.complex128

    def __init__(self):
        super().__init__(name="numpy")

    def __repr__(self) -> str:
        return "BackendNumpy()"

    def abs(self, array: np.ndarray) -> np.ndarray:
        return np.abs(array)

    def any(self, array: np.ndarray) -> np.ndarray:
        return np.any(array)

    def arange(
        self, start: int, limit: Optional[int] = None, delta: int = 1, dtype=np.float64
    ) -> np.ndarray:
        return np.arange(start, limit, delta, dtype=dtype)

    def asnumpy(self, tensor: np.ndarray) -> np.ndarray:
        if isinstance(tensor, np.ndarray):
            return tensor
        return np.array(tensor)

    def assign(self, tensor: np.ndarray, value: np.ndarray) -> np.ndarray:
        tensor = value
        return tensor

    def astensor(self, array: Union[np.ndarray, np.ndarray], dtype=None) -> np.ndarray:
        array = np.array(array)
        return self.cast(array, dtype=dtype or array.dtype)

    def atleast_1d(self, array: np.ndarray, dtype=None) -> np.ndarray:
        return np.atleast_1d(self.astensor(array, dtype))

    def atleast_2d(self, array: np.ndarray, dtype=None) -> np.ndarray:
        return np.atleast_2d(self.astensor(array, dtype))

    def atleast_3d(self, array: np.ndarray, dtype=None) -> np.ndarray:
        array = self.atleast_2d(self.atleast_1d(array))
        if len(array.shape) == 2:
            array = array[None, ...]
        return array

    def block(self, blocks: List[List[np.ndarray]], axes=(-2, -1)) -> np.ndarray:
        rows = [self.concat(row, axis=axes[1]) for row in blocks]
        return self.concat(rows, axis=axes[0])

    def block_diag(self, *blocks: List[np.ndarray]) -> np.ndarray:
        return sp.linalg.block_diag(*blocks)

    def boolean_mask(self, tensor: np.ndarray, mask: np.ndarray) -> np.ndarray:
        return np.array([t for i, t in enumerate(tensor) if mask[i]])

    def cast(self, array: np.ndarray, dtype=None) -> np.ndarray:
        if dtype is None:
            return array
        if dtype not in [self.complex64, self.complex128, "complex64", "complex128"]:
            array = self.real(array)
        return np.array(array, dtype=dtype)

    def clip(self, array, a_min, a_max) -> np.ndarray:
        return np.clip(array, a_min, a_max)

    def concat(self, values: List[np.ndarray], axis: int) -> np.ndarray:
        # tf.concat can concatenate lists of scalars, while np.concatenate errors
        try:
            return np.concatenate(values, axis)
        except ValueError:
            return np.array(values)

    def conj(self, array: np.ndarray) -> np.ndarray:
        return np.conj(array)

    def cos(self, array: np.ndarray) -> np.ndarray:
        return np.cos(array)

    def cosh(self, array: np.ndarray) -> np.ndarray:
        return np.cosh(array)

    def det(self, matrix: np.ndarray) -> np.ndarray:
        return np.linalg.det(matrix)

    def diag(self, array: np.ndarray, k: int = 0) -> np.ndarray:
        if len(array.shape) == 1:
            return np.diag(array, k=k)
        elif len(array.shape) == 2:
            return np.array([np.diag(l, k=k).tolist() for l in array])
        else:
            # fallback into more complex algorithm
            original_sh = array.shape

            ravelled_sh = (np.prod(original_sh[:-1]), original_sh[-1])
            array = array.ravel().reshape(*ravelled_sh)

            ret = []
            for line in array:
                ret.append(np.diag(line, k))

            ret = np.array(ret)
            inner_shape = (
                original_sh[-1] + abs(k),
                original_sh[-1] + abs(k),
            )
            return ret.reshape(original_sh[:-1] + inner_shape)

    def diag_part(self, array: np.ndarray, k: int) -> np.ndarray:
        ret = np.diagonal(array, offset=k, axis1=-2, axis2=-1)
        ret.flags.writeable = True
        return ret

    def set_diag(self, array: np.ndarray, diag: np.ndarray, k: int) -> np.ndarray:
        i = np.arange(0, array.shape[-2] - abs(k))
        if k < 0:
            i -= array.shape[-2] - abs(k)

        j = np.arange(abs(k), array.shape[-1])
        if k < 0:
            j -= abs(k)

        array[..., i, j] = diag

        return array

    def einsum(self, string: str, *tensors) -> Optional[np.ndarray]:
        if type(string) is str:
            return np.einsum(string, *tensors)
        return None  # provide same functionality as numpy.einsum or upgrade to opt_einsum

    def exp(self, array: np.ndarray) -> np.ndarray:
        return np.exp(array)

    def expand_dims(self, array: np.ndarray, axis: int) -> np.ndarray:
        return np.expand_dims(array, axis)

    def expm(self, matrix: np.ndarray) -> np.ndarray:
        return scipy_expm(matrix)

    def eye(self, size: int, dtype=np.float64) -> np.ndarray:
        return np.eye(size, dtype=dtype)

    def eye_like(self, array: np.ndarray) -> np.ndarray:
        return np.eye(array.shape[-1], dtype=array.dtype)

    def from_backend(self, value) -> bool:
        return isinstance(value, np.ndarray)

    def gather(self, array: np.ndarray, indices: np.ndarray, axis: int = 0) -> np.ndarray:
        return np.take(array, indices, axis=axis)

    def imag(self, array: np.ndarray) -> np.ndarray:
        return np.imag(array)

    def inv(self, tensor: np.ndarray) -> np.ndarray:
        return np.linalg.inv(tensor)

    def is_trainable(self, tensor: np.ndarray) -> bool:  # pylint: disable=unused-argument
        return False

    def lgamma(self, x: np.ndarray) -> np.ndarray:
        return np.array([mlgamma(v) for v in x])

    def log(self, x: np.ndarray) -> np.ndarray:
        return np.log(x)

    def make_complex(self, real: np.ndarray, imag: np.ndarray) -> np.ndarray:
        return real + 1j * imag

    @Autocast()
    def matmul(self, *matrices: np.ndarray) -> np.ndarray:
        mat = matrices[0]
        for matrix in matrices[1:]:
            mat = np.matmul(mat, matrix)
        return mat

    @Autocast()
    def matvec(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return self.matmul(a, b[:, None])[:, 0]

    @Autocast()
    def maximum(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.maximum(a, b)

    @Autocast()
    def minimum(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.minimum(a, b)

    def new_variable(
        self,
        value,
        bounds: Union[Tuple[Optional[float], Optional[float]], None],
        name: str,
        dtype=np.float64,
    ):  # pylint: disable=unused-argument
        return np.array(value, dtype=dtype)

    def new_constant(self, value, name: str, dtype=np.float64):  # pylint: disable=unused-argument
        return np.array(value, dtype=dtype)

    def norm(self, array: np.ndarray) -> np.ndarray:
        return np.linalg.norm(array)

    def ones(self, shape: Sequence[int], dtype=np.float64) -> np.ndarray:
        return np.ones(shape, dtype=dtype)

    def ones_like(self, array: np.ndarray) -> np.ndarray:
        return np.ones(array.shape, dtype=array.dtype)

    @Autocast()
    def outer(self, array1: np.ndarray, array2: np.ndarray) -> np.ndarray:
        return np.tensordot(array1, array2, [[], []])

    def pad(
        self,
        array: np.ndarray,
        paddings: Sequence[Tuple[int, int]],
        mode="CONSTANT",
        constant_values=0,
    ) -> np.ndarray:
        if mode == "CONSTANT":
            mode = "constant"
        return np.pad(array, paddings, mode, constant_values=constant_values)

    @staticmethod
    def pinv(matrix: np.ndarray) -> np.ndarray:
        return np.linalg.pinv(matrix)

    @Autocast()
    def pow(self, x: np.ndarray, y: float) -> np.ndarray:
        return np.power(x, y)

    def real(self, array: np.ndarray) -> np.ndarray:
        return np.real(array)

    def reshape(self, array: np.ndarray, shape: Sequence[int]) -> np.ndarray:
        return np.reshape(array, shape)

    def round(self, array: np.ndarray, decimals: int = 0) -> np.ndarray:
        return np.round(array, decimals)

    def sin(self, array: np.ndarray) -> np.ndarray:
        return np.sin(array)

    def sinh(self, array: np.ndarray) -> np.ndarray:
        return np.sinh(array)

    def solve(self, matrix: np.ndarray, rhs: np.ndarray) -> np.ndarray:
        if len(rhs.shape) == len(matrix.shape) - 1:
            rhs = np.expand_dims(rhs, -1)
            return np.linalg.solve(matrix, rhs)[..., 0]
        return np.linalg.solve(matrix, rhs)

    def sqrt(self, x: np.ndarray, dtype=None) -> np.ndarray:
        return np.sqrt(self.cast(x, dtype))

    def sum(self, array: np.ndarray, axes: Sequence[int] = None):
        if axes is None:
            return np.sum(array)

        ret = array
        for axis in axes:
            ret = np.sum(ret, axis=axis)
        return ret

    @Autocast()
    def tensordot(self, a: np.ndarray, b: np.ndarray, axes: List[int]) -> np.ndarray:
        return np.tensordot(a, b, axes)

    def tile(self, array: np.ndarray, repeats: Sequence[int]) -> np.ndarray:
        return np.tile(array, repeats)

    def trace(self, array: np.ndarray, dtype=None) -> np.ndarray:
        return self.cast(np.trace(array), dtype)

    def transpose(self, a: np.ndarray, perm: Sequence[int] = None) -> Optional[np.ndarray]:
        if a is None:
            return None  # TODO: remove and address None inputs where tranpose is used
        return np.transpose(a, axes=perm)

    @Autocast()
    def update_tensor(
        self, tensor: np.ndarray, indices: np.ndarray, values: np.ndarray
    ) -> np.ndarray:
        indices = self.atleast_2d(indices)
        for i, v in zip(indices, values):
            tensor[tuple(i)] = v
        return tensor

    @Autocast()
    def update_add_tensor(
        self, tensor: np.ndarray, indices: np.ndarray, values: np.ndarray
    ) -> np.ndarray:
        indices = self.atleast_2d(indices)
        for i, v in zip(indices, values):
            tensor[tuple(i)] += v
        return tensor

    def zeros(self, shape: Sequence[int], dtype=np.float64) -> np.ndarray:
        return np.zeros(shape, dtype=dtype)

    def zeros_like(self, array: np.ndarray) -> np.ndarray:
        return np.zeros(np.array(array).shape, dtype=array.dtype)

    def map_fn(self, func, elements):
        # Is this done like this?
        return np.array([func(e) for e in elements])

    def squeeze(self, tensor, axis=None):
        return np.squeeze(tensor, axis=axis)

    def cholesky(self, input: np.ndarray):
        return np.linalg.cholesky(input)

    def Categorical(self, probs: np.ndarray, name: str):  # pylint: disable=unused-argument
        class Generator:
            def __init__(self, probs):
                self._probs = probs

            def sample(self):
                idx = [i for i, _ in enumerate(probs)]
                return np.random.choice(idx, p=probs / sum(probs))

        return Generator(probs)

    def MultivariateNormalTriL(self, loc: np.ndarray, scale_tril: np.ndarray):
        class Generator:
            def __init__(self, mean, cov):
                self._mean = mean
                self._cov = cov

            def sample(self, dtype=None):  # pylint: disable=unused-argument
                fn = np.random.default_rng().multivariate_normal
                ret = fn(self._mean, self._cov)
                return ret

            def prob(self, x):
                return multivariate_normal.pdf(x, mean=self._mean, cov=self._cov)

        scale_tril = scale_tril @ np.transpose(scale_tril)
        return Generator(loc, scale_tril)

    @staticmethod
    def eigvals(tensor: np.ndarray) -> np.ndarray:
        return np.linalg.eigvals(tensor)

    @staticmethod
    def xlogy(x: np.ndarray, y: np.ndarray) -> np.ndarray:
        return scipy_xlogy(x, y)

    @staticmethod
    def eigh(tensor: np.ndarray) -> tuple:
        return np.linalg.eigh(tensor)

    def sqrtm(self, tensor: np.ndarray, dtype, rtol=1e-05, atol=1e-08) -> np.ndarray:
        if np.allclose(tensor, 0, rtol=rtol, atol=atol):
            ret = self.zeros_like(tensor)
        else:
            ret = scipy_sqrtm(tensor)

        if dtype is None:
            return self.cast(ret, self.complex128)
        return self.cast(ret, dtype)

    # ~~~~~~~~~~~~~~~~~
    # Special functions
    # ~~~~~~~~~~~~~~~~~

    @staticmethod
    def DefaultEuclideanOptimizer() -> None:
        return None

    def hermite_renormalized(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, shape: Tuple[int]
    ) -> np.ndarray:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx + 1/2*Ax^2)` at zero, where the series has :math:`sqrt(n!)`
        at the denominator rather than :math:`n!`. It computes all the amplitudes within the
        tensor of given shape.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """

        precision_bits = settings.PRECISION_BITS_HERMITE_POLY

        if precision_bits == 128:  # numba
            G = vanilla(tuple(shape), A, B, C)
        else:  # julia (with precision_bits = 512)
            # The following import must come after running "jl = Julia(compiled_modules=False)" in settings.py
            from julia import Main as Main_julia  # pylint: disable=import-outside-toplevel

            A, B, C = (
                np.array(A).astype(np.complex128),
                np.array(B).astype(np.complex128),
                np.array(C).astype(np.complex128),
            )
            G = Main_julia.Vanilla.vanilla(
                A, B, C.item(), np.array(shape, dtype=np.int64), precision_bits
            )

        return G

    def hermite_renormalized_batch(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, shape: Tuple[int]
    ) -> np.ndarray:
        G = vanilla_batch(tuple(shape), A, B, C)
        return G

    def hermite_renormalized_binomial(
        self,
        A: np.ndarray,
        B: np.ndarray,
        C: np.ndarray,
        shape: Tuple[int],
        max_l2: Optional[float],
        global_cutoff: Optional[int],
    ) -> np.ndarray:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx + 1/2*Ax^2)` at zero, where the series has :math:`sqrt(n!)`
        at the denominator rather than :math:`n!`. The computation fills a tensor of given shape
        up to a given L2 norm or global cutoff, whichever applies first. The max_l2 value, if
        not provided, is set to the default value of the AUTOCUTOFF_PROBABILITY setting.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor (local cutoffs).
            max_l2 (float): The maximum squared L2 norm of the tensor.
            global_cutoff (optional int): The global cutoff.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """
        G, _ = binomial(
            tuple(shape),
            A,
            B,
            C,
            max_l2=max_l2 or settings.AUTOCUTOFF_PROBABILITY,
            global_cutoff=global_cutoff or sum(shape) - len(shape) + 1,
        )

        return G

    def reorder_AB_bargmann(self, A: np.ndarray, B: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        r"""In mrmustard.math.numba.compactFock~ dimensions of the Fock representation are ordered like [mode0,mode0,mode1,mode1,...]
        while in mrmustard.physics.bargmann the ordering is [mode0,mode1,...,mode0,mode1,...]. Here we reorder A and B.
        """
        ordering = np.arange(2 * A.shape[0] // 2).reshape(2, -1).T.flatten()
        A = self.gather(A, ordering, axis=1)
        A = self.gather(A, ordering)
        B = self.gather(B, ordering, axis=0)
        return A, B

    def hermite_renormalized_diagonal(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""First, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.numba.compactFock~
        Then, calculate the required renormalized multidimensional Hermite polynomial.
        """
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_diagonal_reorderedAB(A, B, C, cutoffs=cutoffs)

    def hermite_renormalized_diagonal_reorderedAB(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx - Ax^2)` at zero, where the series has :math:`sqrt(n!)` at the
        denominator rather than :math:`n!`. Note the minus sign in front of ``A``.

        Calculates the diagonal of the Fock representation (i.e. the PNR detection probabilities of all modes)
        by applying the recursion relation in a selective manner.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial.
        """
        poly0, _, _, _, _ = hermite_multidimensional_diagonal(A, B, C, cutoffs)

        return poly0

    def hermite_renormalized_diagonal_batch(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""Same as hermite_renormalized_diagonal but works for a batch of different B's."""
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_diagonal_reorderedAB_batch(A, B, C, cutoffs=cutoffs)

    def hermite_renormalized_diagonal_reorderedAB_batch(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""Same as hermite_renormalized_diagonal_reorderedAB but works for a batch of different B's.

        Args:
            A: The A matrix.
            B: The B vectors.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial from different B values.
        """
        poly0, _, _, _, _ = hermite_multidimensional_diagonal_batch(A, B, C, cutoffs)

        return poly0

    def hermite_renormalized_1leftoverMode(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""First, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.numba.compactFock~
        Then, calculate the required renormalized multidimensional Hermite polynomial.
        """
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_1leftoverMode_reorderedAB(A, B, C, cutoffs=cutoffs)

    def hermite_renormalized_1leftoverMode_reorderedAB(
        self, A: np.ndarray, B: np.ndarray, C: np.ndarray, cutoffs: Tuple[int]
    ) -> np.ndarray:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx - Ax^2)` at zero, where the series has :math:`sqrt(n!)` at the
        denominator rather than :math:`n!`. Note the minus sign in front of ``A``.

        Calculates all possible Fock representations of mode 0,
        where all other modes are PNR detected.
        This is done by applying the recursion relation in a selective manner.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial.
        """
        poly0, _, _, _, _ = hermite_multidimensional_1leftoverMode(A, B, C, cutoffs)
        return poly0

    @staticmethod
    def getitem(tensor, *, key):
        value = np.array(tensor)[key]
        return value

    @staticmethod
    def setitem(tensor, value, *, key):
        _tensor = np.array(tensor)
        value = np.array(value)
        _tensor[key] = value

        return _tensor



"""
This module calculates all possible Fock representations of mode 0,where all other modes are PNR detected.
This is done by applying the recursion relation in a selective manner.
"""

import numba
import numpy as np
from numba import int64, njit
from numba.cpython.unsafe.tuple import tuple_setitem

from mrmustard.math.lattice.strategies.compactFock.helperFunctions import (
    SQRT,
    construct_dict_params,
    repeat_twice,
)


@njit
def write_block(
    i, arr_write, write, arr_read_pivot, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode
):  # pragma: no cover
    """
    Apply the recurrence relation to blocks of Fock amplitudes (of shape cutoff_leftoverMode x cutoff_leftoverMode)
    This is the coarse-grained version of applying the recurrence relation of mrmustard.math.compactFock.compactFock_diagonal_amps once.
    """
    m, n = 0, 0
    A_adapted = A[i, 2:]
    G_in_adapted = G_in[0, 0]
    arr_write[(0, 0) + write] = (GB[0, 0, i] + A_adapted @ G_in_adapted) / K_i[i - 2]

    m = 0
    A_adapted = A[i, 1:]
    for n in range(1, cutoff_leftoverMode):
        G_in_adapted = np.hstack(
            (np.array([arr_read_pivot[(0, n - 1) + read_GB] * np.sqrt(n)]), G_in[0, n])
        )
        arr_write[(0, n) + write] = (GB[0, n, i] + A_adapted @ G_in_adapted) / K_i[i - 2]

    n = 0
    A_adapted = np.hstack((np.array([A[i, 0]]), A[i, 2:]))
    for m in range(1, cutoff_leftoverMode):
        G_in_adapted = np.hstack(
            (np.array([arr_read_pivot[(m - 1, 0) + read_GB] * np.sqrt(m)]), G_in[m, 0])
        )
        arr_write[(m, 0) + write] = (GB[m, 0, i] + A_adapted @ G_in_adapted) / K_i[i - 2]

    A_adapted = A[i]
    for m in range(1, cutoff_leftoverMode):
        for n in range(1, cutoff_leftoverMode):
            G_in_adapted = np.hstack(
                (
                    np.array(
                        [
                            arr_read_pivot[(m - 1, n) + read_GB] * np.sqrt(m),
                            arr_read_pivot[(m, n - 1) + read_GB] * np.sqrt(n),
                        ]
                    ),
                    G_in[m, n],
                )
            )
            arr_write[(m, n) + write] = (GB[m, n, i] + A_adapted @ G_in_adapted) / K_i[i - 2]
    return arr_write


@njit
def read_block(
    arr_write, idx_write, arr_read, idx_read_tail, cutoff_leftoverMode
):  # pragma: no cover
    """
    Read the blocks of Fock amplitudes (of shape cutoff_leftoverMode x cutoff_leftoverMode)
    that are required to apply the recurrence relation and write them to G_in
    """
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            arr_write[m, n, idx_write] = arr_read[
                (
                    m,
                    n,
                )
                + idx_read_tail
            ]
    return arr_write


@njit
def use_offDiag_pivot(
    A, B, M, cutoff_leftoverMode, cutoffs_tail, params, d, arr0, arr2, arr1010, arr1001, arr1
):
    """
    Apply recurrence relation for pivot of type [a+1,a,b,b,c,c,...] / [a,a,b+1,b,c,c,...] / [a,a,b,b,c+1,c,...]
    Args:
        A, B (array, Vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of detected modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        d (int): mode index in which the considered Fock amplitude is off diagonal
            e.g. [a,a,b+1,b,c,c,...] --> b is off diagonal --> d=1
        arr0, arr2, arr1010, arr1001, arr1 (array, array, array, array, array): submatrices of the fock representation
    Returns:
        (array, array, array, array, array): updated versions of arr0, arr2, arr1010, arr1001, arr1
    """
    pivot = repeat_twice(params)
    pivot[2 * d] += 1
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, 2 * M), dtype=np.complex128)

    ########## READ ##########
    read_GB = (2 * d,) + params
    GB = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, len(B)), dtype=np.complex128)
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            GB[m, n] = arr1[(m, n) + read_GB] * B

    # Array0
    G_in = read_block(G_in, 2 * d, arr0, params, cutoff_leftoverMode)

    # read from Array2
    if params[d] > 0:
        params_adapted = tuple_setitem(params, d, params[d] - 1)
        G_in = read_block(G_in, 2 * d + 1, arr2, (d,) + params_adapted, cutoff_leftoverMode)

    # read from Array11
    for i in range(d + 1, M):  # i>d
        if params[i] > 0:
            params_adapted = tuple_setitem(params, i, params[i] - 1)
            G_in = read_block(
                G_in, 2 * i, arr1001, (d, i - d - 1) + params_adapted, cutoff_leftoverMode
            )
            G_in = read_block(
                G_in, 2 * i + 1, arr1010, (d, i - d - 1) + params_adapted, cutoff_leftoverMode
            )

    ########## WRITE ##########
    G_in = np.multiply(K_l, G_in)

    # Array0
    params_adapted = tuple_setitem(params, d, params[d] + 1)
    write = params_adapted
    arr0 = write_block(2 * d + 3, arr0, write, arr1, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode)

    # Array2
    if params[d] + 2 < cutoffs_tail[d]:
        write = (d,) + params
        arr2 = write_block(
            2 * d + 2, arr2, write, arr1, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode
        )

    # Array11
    for i in range(d + 1, M):
        if params[i] + 1 < cutoffs_tail[i]:
            write = (d, i - d - 1) + params
            arr1010 = write_block(
                2 * i + 2, arr1010, write, arr1, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode
            )
            arr1001 = write_block(
                2 * i + 3, arr1001, write, arr1, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode
            )

    return arr0, arr2, arr1010, arr1001


@njit
def use_diag_pivot(A, B, M, cutoff_leftoverMode, cutoffs_tail, params, arr0, arr1):
    """
    Apply recurrence relation for pivot of type [a,a,b,b,c,c...]
    Args:
        A, B (array, Vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of detected modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        params (tuple): (a,b,c,...)
        arr0, arr1 (array, array): submatrices of the fock representation
    Returns:
        (array, array): updated versions of arr0, arr1
    """
    pivot = repeat_twice(params)
    K_l = SQRT[pivot]
    K_i = SQRT[pivot + 1]
    G_in = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, 2 * M), dtype=np.complex128)

    ########## READ ##########
    read_GB = params
    GB = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode, len(B)), dtype=np.complex128)
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            GB[m, n] = arr0[(m, n) + read_GB] * B

    # Array1
    for i in range(2 * M):
        if params[i // 2] > 0:
            params_adapted = tuple_setitem(params, i // 2, params[i // 2] - 1)
            G_in = read_block(
                G_in, i, arr1, (i + 1 - 2 * (i % 2),) + params_adapted, cutoff_leftoverMode
            )  # [i+1-2*(i%2) for i in range(6)] == [1,0,3,2,5,4]

    ########## WRITE ##########
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode):
            G_in[m, n] = np.multiply(K_l, G_in[m, n])

    # Array1
    for i in range(2 * M):
        if params[i // 2] + 1 < cutoffs_tail[i // 2]:
            # this if statement prevents a few elements from being written that will never be read
            if i != 1 or params[0] + 2 < cutoffs_tail[0]:
                write = (i,) + params
                arr1 = write_block(
                    i + 2, arr1, write, arr0, read_GB, G_in, GB, A, K_i, cutoff_leftoverMode
                )

    return arr1


@njit
def fock_representation_1leftoverMode_amps_NUMBA(
    A,
    B,
    M,
    cutoff_leftoverMode,
    cutoffs_tail,
    arr0,
    arr2,
    arr1010,
    arr1001,
    arr1,
    tuple_type,
    list_type,
    zero_tuple,
):
    """
    Returns the density matrices in the upper, undetected mode of a circuit when all other modes are PNR detected
    according to algorithm 2 of https://doi.org/10.22331/q-2023-08-29-1097
    Args:
        A, B (array, vector): required input for recurrence relation (given by mrmustard.physics.fock.ABC)
        M (int): number of modes
        cutoffs (tuple): upper bounds for the number of photons in each mode
        arr0 (array): submatrix of the fock representation that contains Fock amplitudes of the type [a,a,b,b,c,c...]
            (!) should already contain G0 at position (0,)*M
        arr2 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+2,a,b,b,c,c...] / [a,a,b+2,b,c,c...] / ...
        arr1010 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b+1,b,c,c,...] / [a+1,a,b,b,c+1,c,...] / [a,a,b+1,b,c+1,c,...] / ...
        arr1001 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b+1,c,c,...] / [a+1,a,b,b,c,c+1,...] / [a,a,b+1,b,c,c+1,...] / ...
        arr1 (array): submatrix of the fock representation that contains Fock amplitudes of the types [a+1,a,b,b,c,c...] / [a,a+1,b,b,c,c...] / [a,a,b+1,b,c,c...] / ...
        tuple_type, list_type (Numba types): numba types that need to be defined outside of numba compiled functions
    Returns:
        Tensor: the fock representation
    """

    # fill first mode for all PNR detections equal to zero
    for m in range(cutoff_leftoverMode - 1):
        arr0[(m + 1, 0) + zero_tuple] = (
            arr0[(m, 0) + zero_tuple] * B[0] + np.sqrt(m) * A[0, 0] * arr0[(m - 1, 0) + zero_tuple]
        ) / np.sqrt(m + 1)
    for m in range(cutoff_leftoverMode):
        for n in range(cutoff_leftoverMode - 1):
            arr0[(m, n + 1) + zero_tuple] = (
                arr0[(m, n) + zero_tuple] * B[1]
                + np.sqrt(m) * A[1, 0] * arr0[(m - 1, n) + zero_tuple]
                + np.sqrt(n) * A[1, 1] * arr0[(m, n - 1) + zero_tuple]
            ) / np.sqrt(n + 1)

    dict_params = construct_dict_params(cutoffs_tail, tuple_type, list_type)
    for sum_params in range(sum(cutoffs_tail)):
        for params in dict_params[sum_params]:
            # diagonal pivots: aa,bb,cc,dd,...
            if (cutoffs_tail[0] == 1) or (params[0] < cutoffs_tail[0] - 1):
                arr1 = use_diag_pivot(
                    A, B, M - 1, cutoff_leftoverMode, cutoffs_tail, params, arr0, arr1
                )
            # off-diagonal pivots: d=0: (a+1)a,bb,cc,dd,... | d=1: 00,(b+1)b,cc,dd | 00,00,(c+1)c,dd | ...
            for d in range(M - 1):
                if np.all(np.array(params)[:d] == 0) and (params[d] < cutoffs_tail[d] - 1):
                    arr0, arr2, arr1010, arr1001 = use_offDiag_pivot(
                        A,
                        B,
                        M - 1,
                        cutoff_leftoverMode,
                        cutoffs_tail,
                        params,
                        d,
                        arr0,
                        arr2,
                        arr1010,
                        arr1001,
                        arr1,
                    )
    return arr0, arr2, arr1010, arr1001, arr1


def fock_representation_1leftoverMode_amps(A, B, G0, M, cutoffs):
    """
    First initialise the submatrices of G (of which the shape depends on cutoff and M)
    and some other constants
    (These initialisations currently cannot be done using Numba.)
    Then calculate the fock representation.
    """

    cutoff_leftoverMode = cutoffs[0]
    cutoffs_tail = tuple(cutoffs[1:])
    tuple_type = numba.types.UniTuple(int64, M - 1)
    list_type = numba.types.ListType(tuple_type)
    zero_tuple = (0,) * (M - 1)

    arr0 = np.zeros((cutoff_leftoverMode, cutoff_leftoverMode) + cutoffs_tail, dtype=np.complex128)
    arr0[(0,) * (M + 1)] = G0
    arr2 = np.zeros(
        (cutoff_leftoverMode, cutoff_leftoverMode) + (M - 1,) + cutoffs_tail, dtype=np.complex128
    )
    arr1 = np.zeros(
        (cutoff_leftoverMode, cutoff_leftoverMode) + (2 * (M - 1),) + cutoffs_tail,
        dtype=np.complex128,
    )
    if M == 2:
        arr1010 = np.zeros((1, 1, 1, 1, 1), dtype=np.complex128)
        arr1001 = np.zeros((1, 1, 1, 1, 1), dtype=np.complex128)
    else:
        arr1010 = np.zeros(
            (cutoff_leftoverMode, cutoff_leftoverMode) + (M - 1, M - 2) + cutoffs_tail,
            dtype=np.complex128,
        )
        arr1001 = np.zeros(
            (cutoff_leftoverMode, cutoff_leftoverMode) + (M - 1, M - 2) + cutoffs_tail,
            dtype=np.complex128,
        )
    return fock_representation_1leftoverMode_amps_NUMBA(
        A,
        B,
        M,
        cutoff_leftoverMode,
        cutoffs_tail,
        arr0,
        arr2,
        arr1010,
        arr1001,
        arr1,
        tuple_type,
        list_type,
        zero_tuple,
    )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains logic for the text-based circuit drawer for MrMustard
"""
from collections import defaultdict

__all__ = [
    "circuit_text",
]


def mode_set(op):
    "includes modes in between min and max of op.modes"
    return set(range(min(op.modes), max(op.modes) + 1))


def drawable_layers(ops):
    r"""Determine non-overlapping yet dense placement of ops into layers for drawing.
    Args:
        ops Iterable[op]: a list of operations

    Returns:
        dict[int:list[op]] : At index k is a list of operations for the k-th layer
    """
    layers = defaultdict(list)
    k = 0
    for new_op in ops:
        # if there's any overlap, add to next layer
        if any(mode_set(new_op) & mode_set(op) for op in layers[k]):
            k += 1
        layers[k].append(new_op)
    return layers


def _add_grouping_symbols(op, layer_str):
    r"""Adds symbols indicating the extent of a given object."""
    S = mode_set(op)
    if len(S) > 1:
        layer_str[min(S)] = "╭"
        layer_str[max(S)] = "╰"
        for w in range(min(S) + 1, max(S)):
            layer_str[w] = "├" if w in op.modes else "│"  # other option: ┼
    return layer_str


def _add_op(op, layer_str, decimals):
    r"""Updates `layer_str` with `op` operation."""
    layer_str = _add_grouping_symbols(op, layer_str)
    control = []
    if op.__class__.__qualname__ in ["BSgate", "MZgate", "CZgate", "CXgate"]:
        control = [op.modes[0]]
    label = op.short_name
    if decimals is not None:
        param_string = op.parameter_set.to_string(decimals)
        if param_string == "":
            param_string = str(len(op.modes))
        label += "(" + param_string + ")"

    for w in op.modes:
        layer_str[w] += "•" if w in control else label

    return layer_str


def circuit_text(
    ops,
    decimals=None,
):
    r"""Text based diagram for a quantum circuit.
    Args:
        ops (List[Transformation]): the operations and measurements to draw as a list of MrMustard operations
        decimals (optional(int)): How many decimal points to include when formatting operation parameters.
            Default ``None`` will omit parameters from operation labels.
    Returns:
        str : String based graphic of the circuit.
    """
    # get all modes used by the ops and sort them
    modes = sorted(list(set().union(*[op.modes for op in ops])))
    # include all modes between min and max (need to draw over them)
    all_modes = range(min(modes), max(modes) + 1)

    totals = [f"{mode}: " for mode in all_modes]
    line_length = max(len(s) for s in totals)
    totals = [s.rjust(line_length, " ") for s in totals]
    filler = "─"

    for layer in drawable_layers(ops).values():
        layer_str = [filler] * (max(all_modes) - min(all_modes) + 1)
        for op in layer:
            layer_str = _add_op(op, layer_str, decimals)

        max_label_len = max(len(s) for s in layer_str)
        layer_str = [s.ljust(max_label_len, filler) for s in layer_str]

        line_length += max_label_len + 1  # one for the filler character

        totals = [filler.join([t, s]) + filler for t, s in zip(totals, layer_str)]

    return "\n".join(totals)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import njit

from mrmustard.math.lattice import paths, steps
from mrmustard.utils.typing import ComplexMatrix, ComplexTensor, ComplexVector
from .flat_indices import first_available_pivot, lower_neighbors, shape_to_strides

__all__ = ["vanilla", "vanilla_batch", "vanilla_jacobian", "vanilla_vjp"]


@njit
def vanilla(shape: tuple[int, ...], A, b, c) -> ComplexTensor:  # pragma: no cover
    r"""Vanilla Fock-Bargmann strategy.

    Flattens the tensors, then fills it by iterating over all indices in the order
    given by ``np.ndindex``. Finally, it reshapes the tensor before returning.

    Args:
        shape (tuple[int, ...]): shape of the output tensor
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): B vector of the Fock-Bargmann representation
        c (complex): vacuum amplitude

    Returns:
        np.ndarray: Fock representation of the Gaussian tensor with shape ``shape``
    """
    # calculate the strides
    strides = shape_to_strides(np.array(shape))

    # init flat output tensor
    ret = np.array([0 + 0j] * np.prod(np.array(shape)))

    # initialize the indeces.
    # ``index`` is the index of the flattened output tensor, while
    # ``index_u_iter`` iterates through the unravelled counterparts of
    # ``index``.
    index = 0
    index_u_iter = np.ndindex(shape)
    next(index_u_iter)

    # write vacuum amplitude
    ret[0] = c

    # iterate over the rest of the indices
    for index_u in index_u_iter:
        # update index
        index += 1

        # calculate pivot's contribution
        i, pivot = first_available_pivot(index, strides)
        value_at_index = b[i] * ret[pivot]

        # add the contribution of pivot's lower's neighbours
        ns = lower_neighbors(pivot, strides, i)
        (j0, n0) = next(ns)
        value_at_index += A[i, j0] * np.sqrt(index_u[j0] - 1) * ret[n0]
        for j, n in ns:
            value_at_index += A[i, j] * np.sqrt(index_u[j]) * ret[n]
        ret[index] = value_at_index / np.sqrt(index_u[i])

    return ret.reshape(shape)


@njit
def vanilla_batch(shape: tuple[int, ...], A, b, c) -> ComplexTensor:  # pragma: no cover
    r"""Vanilla Fock-Bargmann strategy for batched ``b``, with batched dimension on the
    last index.

    Fills the tensor by iterating over all indices in the order given by ``np.ndindex``.

    Args:
        shape (tuple[int, ...]): shape of the output tensor with the batch dimension on the last term
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): batched B vector of the Fock-Bargmann representation, the batch dimension is on the last index
        c (complex): vacuum amplitude

    Returns:
        np.ndarray: Fock representation of the Gaussian tensor with shape ``shape``
    """

    # init output tensor
    G = np.zeros(shape, dtype=np.complex128)

    # initialize path iterator
    path = np.ndindex(shape[:-1])  # We know the last dimension is the batch one

    # write vacuum amplitude
    G[next(path)] = c

    # iterate over the rest of the indices
    for index in path:
        G[index] = steps.vanilla_step_batch(G, A, b, index)

    return G


@njit
def vanilla_jacobian(
    G, A, b, c
) -> tuple[ComplexTensor, ComplexTensor, ComplexTensor]:  # pragma: no cover
    r"""Vanilla Fock-Bargmann strategy gradient. Returns dG/dA, dG/db, dG/dc.
    Notice that G is a holomorphic function of A, b, c. This means that there is only
    one gradient to care about for each parameter (i.e. not dG/dA.conj() etc).
    """

    # init output tensors
    dGdA = np.zeros(G.shape + A.shape, dtype=np.complex128)
    dGdb = np.zeros(G.shape + b.shape, dtype=np.complex128)
    dGdc = G / c

    # initialize path iterator
    path = paths.ndindex_path(G.shape)

    # skip first index
    next(path)

    # iterate over the rest of the indices
    for index in path:
        dGdA, dGdb = steps.vanilla_step_jacobian(G, A, b, index, dGdA, dGdb)

    return dGdA, dGdb, dGdc


@njit
def vanilla_vjp(G, c, dLdG) -> tuple[ComplexMatrix, ComplexVector, complex]:  # pragma: no cover
    r"""Vanilla Fock-Bargmann strategy gradient. Returns dL/dA, dL/db, dL/dc.

    Args:
        G (np.ndarray): Tensor result of the forward pass
        c (complex): vacuum amplitude
        dLdG (np.ndarray): gradient of the loss with respect to the output tensor

    Returns:
        tuple[np.ndarray, np.ndarray, complex]: dL/dA, dL/db, dL/dc
    """
    shape = G.shape

    # calculate the strides
    strides = shape_to_strides(np.array(shape))

    # linearize G
    G_lin = G.flatten()

    # init gradients
    D = G.ndim
    dA = np.zeros((D, D), dtype=np.complex128)  # component of dL/dA
    db = np.zeros(D, dtype=np.complex128)  # component of dL/db
    dLdA = np.zeros_like(dA)
    dLdb = np.zeros_like(db)

    # initialize the indices.
    # ``index`` is the index of the flattened output tensor, while
    # ``index_u_iter`` iterates through the unravelled counterparts of
    # ``index``.
    index = 0
    index_u_iter = np.ndindex(shape)
    next(index_u_iter)

    for index_u in index_u_iter:
        index += 1

        ns = lower_neighbors(index, strides, 0)

        for i, _ in enumerate(db):
            _, n = next(ns)
            db[i] = np.sqrt(index_u[i]) * G_lin[n]
            dA[i, i] = 0.5 * np.sqrt(index_u[i] * (index_u[i] - 1)) * G_lin[n - strides[i]]
            for j in range(i + 1, len(db)):
                dA[i, j] = np.sqrt(index_u[i] * index_u[j]) * G_lin[n - strides[j]]

        dLdA += dA * dLdG[index_u]
        dLdb += db * dLdG[index_u]

    dLdc = np.sum(G_lin.reshape(shape) * dLdG) / c

    return dLdA, dLdb, dLdc



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the logic for cachin tensor functions in Mr Mustard."""

from functools import lru_cache, wraps
from mrmustard.math.backend_manager import BackendManager
import numpy as np

math = BackendManager()


def tensor_int_cache(fn):
    """Decorator function to cache functions with a 1D Tensor (Vector) and int as arguments,
    that is, functions with signature ``func(x: Vector, n: int)``.

    To do so the input vector (non-hashable) is converted into a numpy array (non-hashable)
    and then into a tuple (hashable). This tuple is used by ``functools.lru_cache`` to cache
    the result."""

    @lru_cache
    def cached_wrapper(hashable_array, cutoff):
        array = np.array(hashable_array, dtype=np.float64)
        return fn(array, cutoff)

    @wraps(fn)
    def wrapper(tensor, cutoff):
        return cached_wrapper(tuple(math.asnumpy(tensor)), cutoff)

    # copy lru_cache attributes over too
    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper



"""
This module contains helper functions that are used in
diagonal_amps.py, diagonal_grad.py, singleLeftoverMode_amps.py and singleLeftoverMode_grad.py
to validate the input provided by the user.
"""

from typing import Iterable
import numpy as np
from mrmustard.math.lattice.strategies.compactFock.diagonal_amps import (
    fock_representation_diagonal_amps,
)
from mrmustard.math.lattice.strategies.compactFock.diagonal_grad import (
    fock_representation_diagonal_grad,
)
from mrmustard.math.lattice.strategies.compactFock.singleLeftoverMode_amps import (
    fock_representation_1leftoverMode_amps,
)
from mrmustard.math.lattice.strategies.compactFock.singleLeftoverMode_grad import (
    fock_representation_1leftoverMode_grad,
)
from thewalrus._hafnian import input_validation


def hermite_multidimensional_diagonal(A, B, G0, cutoffs, rtol=1e-05, atol=1e-08):
    """
    Validation of user input for mrmustard.math.backend_tensorflow.hermite_renormalized_diagonal
    """
    input_validation(A, atol=atol, rtol=rtol)
    if A.shape[0] != B.shape[0]:
        raise ValueError("The matrix A and vector B have incompatible dimensions")
    if isinstance(cutoffs, Iterable):
        cutoffs = tuple(cutoffs)
    else:
        raise ValueError("cutoffs should be array like of length M")
    M = len(cutoffs)
    if A.shape[0] // 2 != M:
        raise ValueError("The matrix A and cutoffs have incompatible dimensions")
    return fock_representation_diagonal_amps(A, B, G0, M, cutoffs)


def grad_hermite_multidimensional_diagonal(A, B, G0, arr0, arr2, arr1010, arr1001, arr1):
    """
    Validation of user input for gradients of mrmustard.math.backend_tensorflow.hermite_renormalized_diagonal
    """
    if A.shape[0] != B.shape[0]:
        raise ValueError("The matrix A and vector B have incompatible dimensions")
    M = A.shape[0] // 2
    arr0_dA, arr0_dB = fock_representation_diagonal_grad(
        A, B, M, arr0, arr2, arr1010, arr1001, arr1
    )
    arr0_dG0 = np.array(arr0 / G0).astype(np.complex128)
    return arr0_dG0, arr0_dA, arr0_dB


def hermite_multidimensional_1leftoverMode(A, B, G0, cutoffs, rtol=1e-05, atol=1e-08):
    """
    Validation of user input for mrmustard.math.backend_tensorflow.hermite_renormalized_1leftoverMode
    """
    input_validation(A, atol=atol, rtol=rtol)
    if A.shape[0] != B.shape[0]:
        raise ValueError("The matrix A and vector B have incompatible dimensions")
    if isinstance(cutoffs, Iterable):
        cutoffs = tuple(cutoffs)
    else:
        raise ValueError("cutoffs should be array like of length M")
    M = len(cutoffs)
    if A.shape[0] // 2 != M:
        raise ValueError("The matrix A and cutoffs have incompatible dimensions")
    if M <= 1:
        raise ValueError("The number of modes should be greater than 1.")
    return fock_representation_1leftoverMode_amps(A, B, G0, M, cutoffs)


def grad_hermite_multidimensional_1leftoverMode(A, B, G0, arr0, arr2, arr1010, arr1001, arr1):
    """
    Validation of user input for gradients of mrmustard.math.backend_tensorflow.hermite_renormalized_1leftoverMode
    """
    if A.shape[0] != B.shape[0]:
        raise ValueError("The matrix A and vector B have incompatible dimensions")
    M = A.shape[0] // 2
    if M <= 1:
        raise ValueError("The number of modes should be greater than 1.")
    arr0_dA, arr0_dB = fock_representation_1leftoverMode_grad(
        A, B, M, arr0, arr2, arr1010, arr1001, arr1
    )
    arr0_dG0 = np.array(arr0 / G0).astype(np.complex128)
    return arr0_dG0, arr0_dA, arr0_dB


def hermite_multidimensional_diagonal_batch(A, B, G0, cutoffs, rtol=1e-05, atol=1e-08):
    """
    Validation of user input for mrmustard.math.backend_tensorflow.hermite_renormalized_diagonal_batch
    """
    input_validation(A, atol=atol, rtol=rtol)
    if len(B.shape) != 2:
        raise ValueError("B should be two dimensional (vector and batch dimension)")
    if A.shape[0] != B.shape[0]:
        raise ValueError("The matrix A and vector B have incompatible dimensions")
    if isinstance(cutoffs, Iterable):
        cutoffs = tuple(cutoffs)
    else:
        raise ValueError("cutoffs should be array like of length M")
    M = len(cutoffs)
    if A.shape[0] // 2 != M:
        raise ValueError("The matrix A and cutoffs have incompatible dimensions")
    return fock_representation_diagonal_amps(A, B, G0, M, cutoffs)



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TODO: document this module
"""

from typing import Tuple, Sequence
from mrmustard.utils.typing import Tensor

from mrmustard import math
from .parameter import Trainable


def update_symplectic(grads_and_vars: Sequence[Tuple[Tensor, Trainable]], symplectic_lr: float):
    r"""Updates the symplectic parameters using the given symplectic gradients.
    Implemented from:
        Wang J, Sun H, Fiori S. A Riemannian-steepest-descent approach
        for optimization on the real symplectic group.
        Mathematical Methods in the Applied Sciences. 2018 Jul 30;41(11):4273-86.
    """
    for dS_euclidean, S in grads_and_vars:
        Y = math.euclidean_to_symplectic(S, dS_euclidean)
        YT = math.transpose(Y)
        new_value = math.matmul(
            S, math.expm(-symplectic_lr * YT) @ math.expm(-symplectic_lr * (Y - YT))
        )
        math.assign(S, new_value)


def update_orthogonal(grads_and_vars: Sequence[Tuple[Tensor, Trainable]], orthogonal_lr: float):
    r"""Updates the orthogonal parameters using the given orthogonal gradients.
    Implemented from:
        Y Yao, F Miatto, N Quesada - arXiv preprint arXiv:2209.06069, 2022.
    """
    for dO_euclidean, O in grads_and_vars:
        Y = math.euclidean_to_unitary(O, math.real(dO_euclidean))
        new_value = math.matmul(O, math.expm(-orthogonal_lr * Y))
        math.assign(O, new_value)


def update_unitary(grads_and_vars: Sequence[Tuple[Tensor, Trainable]], unitary_lr: float):
    r"""Updates the unitary parameters using the given unitary gradients.
    Implemented from:
        Y Yao, F Miatto, N Quesada - arXiv preprint arXiv:2209.06069, 2022.
    """
    for dU_euclidean, U in grads_and_vars:
        Y = math.euclidean_to_unitary(U, dU_euclidean)
        new_value = math.matmul(U, math.expm(-unitary_lr * Y))
        math.assign(U, new_value)


def update_euclidean(grads_and_vars: Sequence[Tuple[Tensor, Trainable]], euclidean_lr: float):
    """Updates the parameters using the euclidian gradients."""
    math.euclidean_opt.lr = euclidean_lr
    math.euclidean_opt.apply_gradients(grads_and_vars)


# This dictionary relates each Trainable type to an updater function
param_update_method = {
    "euclidean": update_euclidean,
    "symplectic": update_symplectic,
    "unitary": update_unitary,
    "orthogonal": update_orthogonal,
}



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The optimizer module contains all logic for parameter and circuit optimization
in Mr Mustard.

The :class:`Optimizer` uses Adam underneath the hood for Euclidean parameters and
a custom Symplectic optimizer for Gaussian gates and states and an Orthogonal
optimizer for interferometers.

We can turn any simulation in Mr Mustard into an optimization by marking which
parameters we wish to be trainable. Let's take a simple example:
Finding the optimal beamsplitter transmission to get Hong-Ou-Mandel dip on the
many-photon setting.

.. code-block::

    import numpy as np

    from mrmustard import math
    from mrmustard.lab.gates import S2gate, BSgate
    from mrmustard.lab.states import Vacuum
    from mrmustard.lab.circuit import Circuit
    from mrmustard.training import Optimizer

    r = np.arcsinh(1.0)
    s2_0 = S2gate(r=r, phi=0.0, phi_trainable=True)[0, 1]
    s2_1 = S2gate(r=r, phi=0.0, phi_trainable=True)[2, 3]
    bs = BSgate(
            theta=np.arccos(np.sqrt(k / (i + k))) + 0.1 * np.random.normal(),
            phi=np.random.normal(),
            theta_trainable=True,
            phi_trainable=True,
        )[1, 2]

    circ = Circuit([s2_0, s2_1, bs])
    state_in = Vacuum(num_modes=4)

    i, k = 2, 2
    cutoff = 1 + i + k

    def cost_fn():
        return math.abs((state_in >> circ).ket(cutoffs=[cutoff] * 4)[i, 1, i + k - 1, k]) ** 2

    opt = Optimizer(euclidean_lr=0.01)
    opt.minimize(cost_fn, by_optimizing=[circ], max_steps=300)

Then, we can see the optimized value of the parameters, for example

    .. code-block::

        np.cos(bs.theta.value) ** 2

"""

from .optimizer import Optimizer
from .callbacks import TensorboardCallback



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing classes and methods for progress bars."""

from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn
from mrmustard import settings


# pylint: disable=disallowed-name
class ProgressBar:
    "A spiffy loading bar to display the progress during an optimization."

    def __init__(self, max_steps: int):
        self.taskID = None
        if max_steps == 0:
            self.bar = Progress(
                TextColumn("Step {task.completed}/∞"),
                BarColumn(),
                TextColumn("Cost = {task.fields[loss]:.5f}"),
            )
        else:
            self.bar = Progress(
                TextColumn("Step {task.completed}/{task.total} | {task.fields[speed]:.1f} it/s"),
                BarColumn(),
                TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                TextColumn("Cost = {task.fields[loss]:.5f} | ⏳ "),
                TimeRemainingColumn(),
            )
        self.taskID = self.bar.add_task(
            description="Optimizing...",
            start=max_steps > 0,
            speed=0.0,
            total=max_steps,
            loss=1.0,
            refresh=True,
            visible=settings.PROGRESSBAR,
        )

    def step(self, loss):
        """Update bar step and the loss information associated with it."""
        speed = self.bar.tasks[0].speed or 0.0
        self.bar.update(self.taskID, advance=1, refresh=True, speed=speed, loss=loss)

    def __enter__(self):
        return self.bar.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.bar.__exit__(exc_type, exc_val, exc_tb)
